!+
! Module: plotutils_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines to allow convenient test plots to be made.
! 
! Comments: TODO:
!	- put in doco that style etc opt pars should be 255 char or less. (and checks, trims, -> warnings if optional char args have len greater than this.)
!	- can check for NaNs??
!	- return with error if pgopen returns <= 0.
!	- offset status values returned by getClrLimits etc calls.
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module plotutils_mod

  use dal
  use param  
  use errorHandling
  use mathutils_mod
use sort_mod ! only required for mySort().

  implicit none 

  private 

!  public initTrMatrix
  public plotGraph
!  public plotMultiGraph
  public plotHistogram
  public plotImage
  public plotColour3d
  public editColourMap
  public closePgDevice
  public plotGTIs
public test
!  public mouse
  public rotatingDisplay
  public stereoDisplay

  interface plotGraph
    module procedure plotSingleGraph, plotManyGraphs
  end interface

  interface plotHistogram
    module procedure plotHistogramReal, plotHistogramInt
  end interface

  interface plotImage
    module procedure plotImageReal, plotImageDouble, plotImageInt&
    , plotImageLogical
  end interface

  interface rotatingDisplay
    module procedure rotatingDisplayVector, rotatingDisplayArray
  end interface

  interface stereoDisplay
    module procedure stereoDisplayVector, stereoDisplayArray
  end interface

  real, dimension(6) ::&
    trMatrix
  real(kind(0d0)), parameter, public ::&
    myPi = 3.1415926535897931,&
    myTwoPi = 2.0d0 * myPi,&
    myPiOn2 = myPi / 2.0d0,&
    myPiOn180 = myPi / 180.0d0,&
!    myRootPi = 1.7724538509055159,&
!    myRoot2 = 1.4142135623730951,&
!    myRoot3 = 1.7320508075688772,&
!    myLog2 = 0.6931471805599453,&
    myLog10 = 2.3025850929940459

  character(3), parameter ::&
    defaultDevice = '/xs',&
    cmapModDevice = defaultDevice
  integer :: minClrIndex, maxClrIndex, numClrIndices
  real :: realNumCIminus1, realNumCI, oldXedges(1:256), newXedges(1:256)
  integer, parameter ::&
    numReservedClrs = 3,&
    red = 1,&
    grn = 2,&
    blu = 3
  type colourType
    real, dimension(red:blu) :: hue
  end type colourType
  type(colourType) :: colour(0:255), deviceColour(0:255)
  logical :: coloursAlreadyLoaded = .false.

  type obsType
    real :: location(3), locRadius, locTheta, locPhi, viewFrame(3, 3), viewTheta, viewPhi, viewTwist, projOffsets(2), projScale
  end type obsType
  type(obsType) :: defaultObs
!  type(obsType), parameter ::&
!    defaultObs = obsType(0.0, 0.0, 0.0, 6.0, 0.6, 0.4)

!**** set default obs in init?? Or just here? Don't know how to make a parameter of defined type.

  type pointType
    real :: solid(3), proj(2)
  end type pointType

  ! editColourMap variables:
  integer, parameter :: numBtns = 6
  real, parameter ::&
    clrStripMinX = 0.0,&
    clrStripMaxX = 0.05,&
    clrStripMinY = 0.02,&
    clrStripMaxY = 0.98,&
    boxRowMinX = 0.08,&
    boxRowMaxX = 0.84,&
    boxRowMinY = clrStripMinY,&
    boxRowMaxY = clrStripMaxY,&
    boxPadFraction = 0.1,&
    btnColMinX = 0.87,&
    btnColMaxX = 1.0,&
    btnColMinY = 0.45,&
    btnColMaxY = clrStripMaxY,&
    btnPadFraction = 0.15
  real, dimension(red:blu) ::&
    boxMinX, boxMaxX
  real, dimension(numBtns) ::&
    btnMinY, btnMaxY
  character(len=20), dimension(numBtns) :: btnText
  integer :: minClrIndex4ECM, maxClrIndex4ECM, numClrIndices4ECM, selectedCi = -1, selectedHi = -1
  real :: realNumCIminus14ECM, realNumCI4ECM
  type pinType
    logical :: isPinned(red:blu)
    real :: hue(red:blu)
  end type pinType
  type(pinType) :: pins(0:255)
  integer, parameter :: pinSymbol = 0, selectionSymbol = 16, ciBoundary = 10
!  logical :: somePinIsSelected

  type, public :: gtiType
    real(kind(0d0)) :: time
    logical         :: isStart
    integer         :: fileNum
  end type gtiType

logical, parameter :: testFlag = .false.

!  private 
!
!!  public initTrMatrix
!  public plotGraph
!!  public plotMultiGraph
!  public plotHistogram
!  public plotImage
!  public plotColour3d
!  public editColourMap
!  public closePgDevice
!  public plotGTIs
!public test
!!  public mouse
!  public rotatingDisplay
!  public stereoDisplay
!
!!public myLog10

contains 

  !---------------------------------------------------------------------
  subroutine test
  !---------------------------------------------------------------------

    implicit none

    integer :: deviceId, pgopen

    !-------------------------------------------------------------------
    ! Start

deviceId = pgopen('/xs')
call pgclos

  end subroutine test

  !---------------------------------------------------------------------
  subroutine initTrMatrix
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    trMatrix(1) = 0.0
    trMatrix(2) = 1.0
    trMatrix(3) = 0.0
    trMatrix(4) = 0.0
    trMatrix(5) = 0.0
    trMatrix(6) = 1.0

  end subroutine initTrMatrix

  !---------------------------------------------------------------------
  subroutine initDefaultObs
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    defaultObs%locRadius = 6.0
    defaultObs%locTheta = 0.6
    defaultObs%locPhi = 0.4

  end subroutine initDefaultObs

  !---------------------------------------------------------------------
  function plotSingleGraph(y, x, yErrors, yIsNull, suppressYzero, device&
  , rainbow, style, xLabel, yLabel, label, wait, deviceId) result(status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in)              :: y(:)
    real,         intent(in),    optional :: x(:),&
                                             yErrors(:)
    character(*), intent(in),    optional :: device,&
                                             style,&
                                             xLabel,&
                                             yLabel,&
                                             label
    logical,      intent(in),    optional :: yIsNull(:),&
                                             suppressYzero,&
                                             rainbow,&
                                             wait
    integer,      intent(inout), optional :: deviceId

    integer :: status

    integer :: i, vectorSize, localDevId
    real, dimension(:), allocatable :: localX, localYerrors
    logical, dimension(:), allocatable :: localYisNull
    character(255) :: localDevice, localStyle, localXlabel, localYlabel,&
      localLabel
    logical :: localSuppressYzero, withYerrors, localRainbow, localWait
    real :: localY(1:size(y), 1)

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    vectorSize = size(y)
    localY(1:vectorSize, 1) = y

    if (present(x)) then
      allocate(localX(size(x)))
      localX = x
    else
      allocate(localX(vectorSize))
      do i = 1, vectorSize
        localX(i) = real(i)
      end do
    end if

    if (present(yErrors)) then
      allocate(localYerrors(size(yErrors)))
      localYerrors = yErrors
      withYerrors = .true.
    else
      allocate(localYerrors(vectorSize))
      localYerrors = 0.0
      withYerrors = .false.
    end if

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then
      localStyle = trim(style)
    else
      localStyle = 'line'
    end if

    if (present(xLabel)) then
      localXlabel = trim(xLabel)
    else
      localXlabel = ''
    end if

    if (present(yLabel)) then
      localYlabel = trim(yLabel)
    else
      localYlabel = ''
    end if

    if (present(label)) then
      localLabel = trim(label)
    else
      localLabel = ''
    end if

    if (present(yIsNull)) then
      allocate(localYisNull(size(yIsNull)))
      localYisNull = yIsNull
    else
      allocate(localYisNull(vectorSize))
      localYisNull = .false.
    end if

    if (present(suppressYzero)) then
      localSuppressYzero = suppressYzero
    else
      localSuppressYzero = .false.
    end if

    if (present(rainbow)) then ! suppresses compiler warnings.
    end if
    localRainbow = .false.

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(deviceId)) then
      localDevId = deviceId
    else
      localDevId = 0 ! not a valid number - this signals to kernel that a
      ! new device is required.
    end if

    status = plotGraphKernel(localY, localX, withYerrors, localYerrors&
    , localYisNull, localSuppressYzero, localRainbow, localDevice, localStyle&
    , localXlabel, localYlabel, localLabel, localWait, localDevId)

    if (present(deviceId)) then
      deviceId = localDevId
    end if

  end function plotSingleGraph

  !---------------------------------------------------------------------
  function plotManyGraphs(y, x, yErrors, yIsNull, suppressYzero, device&
  , rainbow, style, xLabel, yLabel, label, wait, deviceId) result(status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in)              :: y(:,:)
    real,         intent(in),    optional :: x(:),&
                                             yErrors(:)
    character(*), intent(in),    optional :: device,&
                                             style,&
                                             xLabel,&
                                             yLabel,&
                                             label
    logical,      intent(in),    optional :: yIsNull(:),&
                                             suppressYzero,&
                                             rainbow,&
                                             wait
    integer,      intent(inout), optional :: deviceId
    integer :: status

    integer :: i, vectorSize, localDevId
    real, dimension(:), allocatable :: localX, localYerrors
    logical, dimension(:), allocatable :: localYisNull
    character(255) :: localDevice, localStyle, localXlabel, localYlabel,&
      localLabel
    logical :: localSuppressYzero, withYerrors, localRainbow,&
      localWait

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    vectorSize = size(y, 1)

    if (present(x)) then
      allocate(localX(size(x)))
      localX = x
    else
      allocate(localX(vectorSize))
      do i = 1, vectorSize
        localX(i) = real(i)
      end do
    end if

    if (present(yErrors)) then ! suppresses compiler warnings.
    end if
    allocate(localYerrors(vectorSize))
    localYerrors = 0.0
    withYerrors = .false.

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then ! suppresses compiler warnings.
    end if
    localStyle = 'line'

    if (present(xLabel)) then
      localXlabel = trim(xLabel)
    else
      localXlabel = ''
    end if

    if (present(yLabel)) then
      localYlabel = trim(yLabel)
    else
      localYlabel = ''
    end if

    if (present(label)) then
      localLabel = trim(label)
    else
      localLabel = ''
    end if

    if (present(yIsNull)) then ! suppresses compiler warnings.
    end if
    allocate(localYisNull(vectorSize))
    localYisNull = .false.

    if (present(suppressYzero)) then
      localSuppressYzero = suppressYzero
    else
      localSuppressYzero = .false.
    end if

    if (present(rainbow)) then
      localRainbow = rainbow
    else
      localRainbow = .false.
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(deviceId)) then
      localDevId = deviceId
    else
      localDevId = 0 ! not a valid number - this signals to kernel that a
      ! new device is required.
    end if

    status = plotGraphKernel(y, localX, withYerrors, localYerrors&
    , localYisNull, localSuppressYzero, localRainbow, localDevice, localStyle&
    , localXlabel, localYlabel, localLabel, localWait, localDevId)

    if (present(deviceId)) then
      deviceId = localDevId
    end if

  end function plotManyGraphs

  !---------------------------------------------------------------------
  function plotGraphKernel(y, x, withYerrors, yErrors, yIsNull&
  , suppressYzero, multiColoured, device, style, xLabel, yLabel, label&
  , wait, deviceId) result(status)
  !---------------------------------------------------------------------
!**** there is a potential problem with the device ids which could arise as follows: say the user has already opened a pg device, deviceId=a. He wants to open another, leave it open for further plots. So he calls plotGraph with the deviceId argument present (because its output value is required) but not initialised. If by chance the value happens to be 'a', then the previous plot will be overdrawn rather than a new one created. This can be avoided by the user initialising deviceId to zero or less. This should be documented.

    implicit none

    real, intent(in) :: x(:), y(:,:), yErrors(:)
    character(*), intent(in) :: device, style, xLabel, yLabel, label
    logical, intent(in) :: withYerrors, yIsNull(:), suppressYzero,&
      multiColoured, wait
    integer, intent(inout) :: deviceId
    integer :: status

    integer :: pgopen, i, firstNonNull, iStart, iFinis,&
      vectorSize, numPlots, plotNum, j, deviceStatus,&
      plotCIs(size(y, 2)), nullCI, previousDevId, answerLength
    real :: minX, maxX, minY, maxY, xEdgeWidth, yEdgeWidth,&
      frac, redVal, grnVal, bluVal
    logical :: allYnulls, inNullPatch, openNewDevice
character(10) :: pgplotState

    !-------------------------------------------------------------------
    ! Start:

!print *, 'in plotGraph'

    status = 0 ! default

!    openNewDevice = .true.
!    if (deviceId .ne. 0) then
!      call pgqinf('STATE', pgplotState, answerLength)
!      if (trim(pgplotState) .eq. 'OPEN') then
!        call pgqid(previousDevId)
!        if (deviceId .eq. previousDevId) then ! device is already open.
!          openNewDevice = .false.
!        else
!          call pgslct(deviceId)
!          call pgqid(deviceId)
!          if (deviceId .ne. previousDevId) then ! pgslct succeeded, deviceId
!          ! is now open.
!            openNewDevice = .false.
!          end if
!        end if
!      end if
!    end if

deviceStatus = queryDeviceStatus(deviceId)
openNewDevice = .false.
if (deviceStatus .le. 1) then
  openNewDevice = .true.
else if (deviceStatus .eq. 2) then
  call pgslct(deviceId)
end if

    vectorSize = size(y, 1)
    numPlots   = size(y, 2)

    if (size(x) .lt. vectorSize) then
      status = -1
  return
    else if (size(x) .gt. vectorSize) then
      status = status + 1
    end if
    if (size(yErrors) .lt. vectorSize) then
      status = -2
  return
    else if (size(yErrors) .gt. vectorSize) then
      status = status + 2
    end if
    if (size(yIsNull) .lt. vectorSize) then
      status = -3
  return
    else if (size(yIsNull) .gt. vectorSize) then
      status = status + 4
    end if

    allYnulls = .true.
    do i = 1, vectorSize
      if (.not. yIsNull(i)) then
        allYnulls = .false.
        firstNonNull = i
    exit
      end if
    end do
    if (allYnulls) then
      status = status + 8
    end if

    minX = minval(x)
    maxX = maxval(x)

    if (allYnulls) then
      minY = 0.0
      maxY = 1.0

    else
      if (numPlots .gt. 1) then
        minY = minval(y)
        maxY = maxval(y)

      else if (withYerrors) then
        minY = y(firstNonNull, 1) - yErrors(firstNonNull) 
        maxY = y(firstNonNull, 1) + yErrors(firstNonNull)
        do i = firstNonNull + 1, vectorSize
          if (.not. yIsNull(i)) then
            minY = min(minY, y(i, 1) - yErrors(i))
            maxY = max(maxY, y(i, 1) + yErrors(i))
          end if
        end do

      else
        minY = y(firstNonNull, 1)
        maxY = y(firstNonNull, 1)
        do i = firstNonNull + 1, vectorSize
          if (.not. yIsNull(i)) then
            minY = min(minY, y(i, 1))
            maxY = max(maxY, y(i, 1))
          end if
        end do
      end if
      if (.not. suppressYzero) then
        minY = min(0.0, minY)
        maxY = max(0.0, maxY)
      end if
    end if ! allYnulls

    if (openNewDevice) then
      deviceId = pgopen(trim(device))
      call pgsvp(0.07, 0.999, 0.1, 0.90)
!***** use the standard global values.
      call pgscr(2, 0.5, 0.5, 0.5)
    end if

    call findClrLimits(status)
  if (status .lt. 0) return
!*** make status numbers consistent.
!print *, minClrIndex, maxClrIndex, numClrIndices
    if (numPlots + 5 .gt. numClrIndices) then
    ! Extra colours needed for (i) nulls, (ii) frame (fg), (iii) bg, (iv) extra plot, (v) axes.
print *, 'max ci = ', maxClrIndex
!**** set a warning status value
      plotCIs = maxClrIndex ! vector op
      nullCI = maxClrIndex
    else
      nullCI = 3
      call pgscr(nullCI, 0.3, 1.0, 0.75)
      if (numPlots .eq. 1 .or. .not. multiColoured) then
        plotCIs = 4
        call pgscr(4, 1.0, 0.95, 0.7)
!***** check that first colour is the same whether many plots or one.

      else
        do plotNum = 1, numPlots
          plotCIs(plotNum) = plotNum + 4
          frac = (plotNum - 1) / real(numPlots - 1)
!          if (frac .lt. 0.4) then
          if (frac .lt. 0.25) then
            redVal = 1.0
!            grnVal = frac / 0.4
            grnVal = frac / 0.25
            bluVal = 0.0
!          else if (frac .lt. 0.6) then
          else if (frac .lt. 0.5) then
!            redVal = 1.0 - (frac - 0.4) / 0.2
            redVal = 1.0 - (frac - 0.25) / 0.25
            grnVal = 1.0
            bluVal = 0.0
!          else if (frac .lt. 0.8) then
          else if (frac .lt. 0.75) then
            redVal = 0.0
!            grnVal = 1.0 - (frac - 0.6) / 0.2
!            bluVal = (frac - 0.6) / 0.2
            grnVal = 1.0 - (frac - 0.5) / 0.25
            bluVal = (frac - 0.5) / 0.25
          else
!            redVal = 0.7 * (frac - 0.8) / 0.2
            redVal = 0.7 * (frac - 0.75) / 0.25
            grnVal = 0.0
!            bluVal = 1.0 - 0.3 * (frac - 0.8) / 0.2
            bluVal = 1.0 - 0.3 * (frac - 0.75) / 0.25
          end if
          call pgscr(plotCIs(plotNum), redVal, grnVal, bluVal)
        end do
      end if
    end if
plotCIs = 1 ! vector op
nullCI = 1

    select case(style)
    case('line', 'lines', 'linear')
      xEdgeWidth = 0.0
      yEdgeWidth = 0.05 * (maxY - minY)

      if (openNewDevice) then
        call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
        , minY - yEdgeWidth, maxY + yEdgeWidth)

        call pgsci(1)
        call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
        call pglab(trim(xLabel), trim(yLabel), trim(label))
        call pgsci(2)
        call pgmove(0.0, minY - yEdgeWidth)
        call pgdraw(0.0, maxY + yEdgeWidth)
        call pgmove(minX - xEdgeWidth, 0.0)
        call pgdraw(maxX + xEdgeWidth, 0.0)
      end if

      if (numPlots .gt. 1) then
        do plotNum = 1, numPlots
!print *, plotNum
          call pgsci(plotCIs(plotNum))
          call pgline(vectorSize, x(1:vectorSize), y(1:vectorSize, plotNum))
        end do

      else
        if (yIsNull(1)) then
          inNullPatch = .true.
        else
          inNullPatch = .false.
        end if
        iStart = 1
        do i = 2, vectorSize
          if (inNullPatch .and. .not. yIsNull(i)) then
            iFinis = i
            call pgsci(nullCI)
            call pgmove(x(iStart), 0.0)
            call pgdraw(x(iFinis), 0.0)
            iStart = i
            inNullPatch = .false.

          else if (.not. inNullPatch .and. yIsNull(i)) then
            iFinis = i
            call pgsci(plotCIs(1))
            if (withYerrors) then
!***** if too many x points, bin up the x values.
              call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
              , y(iStart:iFinis, 1))
              do j = iStart, iFinis
                call pgmove(x(j), y(j, 1) - yErrors(j))
                call pgdraw(x(j), y(j, 1) + yErrors(j))
              end do

            else
              call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
              , y(iStart:iFinis, 1))
            end if
            iStart = i
            inNullPatch = .true.
          end if
        end do
        if (inNullPatch) then
          iFinis = vectorSize
          call pgsci(nullCI)
          call pgmove(x(iStart), 0.0)
          call pgdraw(x(iFinis), 0.0)

        else if (.not. inNullPatch) then
          iFinis = vectorSize
          call pgsci(plotCIs(1))
          if (withYerrors) then
!***** if too many x points, bin up the x values.
            call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
            , y(iStart:iFinis, 1))
            do j = iStart, iFinis
              call pgmove(x(j), y(j, 1) - yErrors(j))
              call pgdraw(x(j), y(j, 1) + yErrors(j))
            end do

          else
!do j = iStart, iFinis
!  print *, j, x(j), y(j, 1)
!end do
            call pgline(iFinis - iStart + 1, x(iStart:iFinis)&
            , y(iStart:iFinis, 1))
          end if
        end if
      end if

    case('point', 'points')
!******* add square, diamond, circle??
      xEdgeWidth = 0.0
      yEdgeWidth = 0.05 * (maxY - minY)

      if (openNewDevice) then
        call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
        , minY - yEdgeWidth, maxY + yEdgeWidth)

        call pgsci(1)
        call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
        call pglab(trim(xLabel), trim(yLabel), trim(label))
        call pgsci(2)
        call pgmove(0.0, minY - yEdgeWidth)
        call pgdraw(0.0, maxY + yEdgeWidth)
        call pgmove(minX - xEdgeWidth, 0.0)
        call pgdraw(maxX + xEdgeWidth, 0.0)
      end if

      if (withYerrors) then
        do i = 1, vectorSize
          if (yIsNull(i)) then
            call pgsci(nullCI)
            call pgpt1(x(i), 0.0, -1)
          else
            call pgsci(plotCIs(1))
            call pgmove(x(i), y(i, 1) - yErrors(i))
            call pgdraw(x(i), y(i, 1) + yErrors(i))
          end if
        end do

      else
        do i = 1, vectorSize
          if (yIsNull(i)) then
            call pgsci(nullCI)
            call pgpt1(x(i), 0.0, -1)
          else
            call pgsci(plotCIs(1))
            call pgpt1(x(i), y(i, 1), -1)
          end if
        end do
      end if

    case default
      status = -4
  return
    end select

    if (wait .and. .not. openNewDevice) then
      call pgslct(previousDevId)
    else if (.not. wait) then
      call pgclos
    end if

  end function plotGraphKernel

  !---------------------------------------------------------------------
  function queryDeviceStatus(deviceId)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: deviceId
    integer :: queryDeviceStatus, answerLength, currentDeviceId, testDeviceId
    character(10) :: pgplotState

    !-------------------------------------------------------------------
    ! Start:

    if (deviceId .le. 0) then
      queryDeviceStatus = -1 ! not a valid device ID integer.

    else
      call pgqinf('STATE', pgplotState, answerLength)
      if (trim(pgplotState) .eq. 'OPEN') then
        call pgqid(currentDeviceId)
        if (deviceId .eq. currentDeviceId) then
          queryDeviceStatus = 3 ! deviceId is open and selected.

        else ! deviceId .ne. currentDeviceId
          call pgslct(deviceId)
          call pgqid(testDeviceId)
          if (testDeviceId .eq. currentDeviceId) then
            queryDeviceStatus = 1 ! pgslct didn't succeed, there are some
            ! pgplot devices open, but deviceId is not one of them.

          else
            call pgslct(currentDeviceId)
            queryDeviceStatus = 2 ! deviceId is open but not currently selected.
          end if
        end if

      else
        queryDeviceStatus = 0 ! No pgplot devices are open.
      end if
    end if

  end function queryDeviceStatus

  !---------------------------------------------------------------------
  function plotMultiGraph(y, device) result(status)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: y(:,:)
    character(*), intent(in), optional :: device
    integer :: status

    integer :: i, vectorSize, numPlots, deviceId, pgopen, plotNum
    real, allocatable :: x(:)
    character(255) :: localDevice
    real :: minX, minY, maxX, maxY, xEdgeWidth, yEdgeWidth

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

    numPlots   = size(y, 1)
    vectorSize = size(y, 2)

    allocate(x(vectorSize))
    do i = 1, vectorSize
      x(i) = real(i - 1)
    end do
    minX = 0.0
    maxX = real(vectorSize - 1)

    minY = minval(y)
    maxY = maxval(y)
    minY = min(0.0, minY)
    maxY = max(0.0, maxY)

    deviceId = pgopen(trim(localDevice))
    call pgsvp(0.07, 0.999, 0.1, 0.90)

    xEdgeWidth = 0.0
    yEdgeWidth = 0.05 * (maxY - minY)
    call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
    , minY - yEdgeWidth, maxY + yEdgeWidth)

    do plotNum = 1, numPlots
      call pgsci(plotNum + 1)
      call pgline(vectorSize, x, y(plotNum, 1:vectorSize))
    end do

    call pgsci(1)
    call pgbox('BCNT', 0., 0, 'ABCNT', 0., 0)
!    call pglab(trim(xLabel), trim(yLabel), trim(label))
    call pgclos

  end function plotMultiGraph

  !---------------------------------------------------------------------
  function plotHistogramReal(binContents, binErrors, binEdges, binIsNull&
  , divideByBinWidth, device, xLabel, yLabel, label) result(status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in)           :: binContents(:)
    real,         intent(in), optional :: binErrors(:)
    real,         intent(in), optional :: binEdges(:)
    logical,      intent(in), optional :: binIsNull(:)
    logical,      intent(in), optional :: divideByBinWidth
    character(*), intent(in), optional :: device,&
                                          xLabel,&
                                          yLabel,&
                                          label
    integer :: status

    real, allocatable :: myBinEdges(:), myBinErrors(:)
    logical, allocatable :: myBinIsNull(:)
    logical :: myDivideByBinWidth, withYerrors
    integer :: i
    character(255) :: localDevice, localXlabel, localYlabel, localLabel

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    if (present(binEdges)) then
      allocate(myBinEdges(size(binEdges)))
      myBinEdges = binEdges
    else
      allocate(myBinEdges(size(binContents) + 1))
      do i = 1, size(binContents) + 1
        myBinEdges(i) = (i - 1) / real(size(binContents))
      end do
    end if

    if (present(binErrors)) then
      allocate(myBinErrors(size(binErrors)))
      myBinErrors = binErrors
      withYerrors = .true.
    else
      allocate(myBinErrors(size(binContents)))
      myBinErrors = 0.0
      withYerrors = .false.
    end if

    if (present(binIsNull)) then
      allocate(myBinIsNull(size(binIsNull)))
      myBinIsNull = binIsNull
    else
      allocate(myBinIsNull(size(binContents)))
      myBinIsNull = .false.
    end if

    if (present(divideByBinWidth)) then
      myDivideByBinWidth = divideByBinWidth
    else
      myDivideByBinWidth = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(xLabel)) then
      localXlabel = trim(xLabel)
    else
      localXlabel = ''
    end if

    if (present(yLabel)) then
      localYlabel = trim(yLabel)
    else
      localYlabel = ''
    end if

    if (present(label)) then
      localLabel = trim(label)
    else
      localLabel = ''
    end if

    status = plotHistogramKernel(binContents, withYerrors&
    , myBinErrors, myBinEdges, myBinIsNull, myDivideByBinWidth, localDevice&
    , localXlabel, localYlabel, localLabel)

  end function plotHistogramReal

  !---------------------------------------------------------------------
  function plotHistogramInt(binContents, binErrors, binEdges, binIsNull&
  , divideByBinWidth, device, xLabel, yLabel, label) result(status)
  !---------------------------------------------------------------------

    implicit none

    integer,      intent(in)           :: binContents(:)
    integer,      intent(in), optional :: binErrors(:)
    real,         intent(in), optional :: binEdges(:)
    logical,      intent(in), optional :: binIsNull(:)
    logical,      intent(in), optional :: divideByBinWidth
    character(*), intent(in), optional :: device,&
                                          xLabel,&
                                          yLabel,&
                                          label
    integer :: status

    real, allocatable :: myBinEdges(:), myBinErrors(:)
    logical, allocatable :: myBinIsNull(:)
    logical :: myDivideByBinWidth, withYerrors
    integer :: i
    character(255) :: localDevice, localXlabel, localYlabel, localLabel

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    if (present(binErrors)) then
      allocate(myBinErrors(size(binErrors)))
      myBinErrors = real(binErrors)
      withYerrors = .true.
    else
      allocate(myBinErrors(size(binContents)))
      myBinErrors = 0.0
      withYerrors = .false.
    end if

    if (present(binEdges)) then
      allocate(myBinEdges(size(binEdges)))
      myBinEdges = binEdges
    else
      allocate(myBinEdges(size(binContents) + 1))
      do i = 1, size(binContents) + 1
        myBinEdges(i) = (i - 1) / real(size(binContents))
      end do
    end if

    if (present(binIsNull)) then
      allocate(myBinIsNull(size(binIsNull)))
      myBinIsNull = binIsNull
    else
      allocate(myBinIsNull(size(binContents)))
      myBinIsNull = .false.
    end if

    if (present(divideByBinWidth)) then
      myDivideByBinWidth = divideByBinWidth
    else
      myDivideByBinWidth = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(xLabel)) then
      localXlabel = trim(xLabel)
    else
      localXlabel = ''
    end if

    if (present(yLabel)) then
      localYlabel = trim(yLabel)
    else
      localYlabel = ''
    end if

    if (present(label)) then
      localLabel = trim(label)
    else
      localLabel = ''
    end if

    status = plotHistogramKernel(real(binContents), withYerrors&
    , myBinErrors, myBinEdges, myBinIsNull, myDivideByBinWidth, localDevice&
    , localXlabel, localYlabel, localLabel)

  end function plotHistogramInt

  !---------------------------------------------------------------------
  function plotHistogramKernel(binContents, withYerrors, binErrors, binEdges&
  , binIsNull, divideByBinWidth, device, xLabel, yLabel, label) result(status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in) :: binContents(:)
    real,         intent(in) :: binErrors(:)
    real,         intent(in) :: binEdges(:)
    logical,      intent(in) :: binIsNull(:)
    logical,      intent(in) :: withYerrors,&
                                divideByBinWidth
    character(*), intent(in) :: device,&
                                xLabel,&
                                yLabel,&
                                label
    integer :: status

    real, allocatable :: myBinEdges(:), myBinContents(:), myBinErrors(:)
    integer :: hsize, deviceID, bin, pgopen, oldBin, newBin, newNumBins, i
    real :: minx, maxx, miny, maxy, xEdgeWidth, yEdgeWidth, minBinWidth
    real :: leftEdgePixels, riteEdgePixels, botmEdgePixels, toppEdgePixels
    logical :: someNegativeErrors, lastBinWasSmall, someBinsWereAmalgamated

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default

    hsize = size(binContents)

    if (size(binErrors) .lt. hsize) then
      status = -1
  return
    else if (size(binErrors) .gt. hsize) then
      status = status + 1
    end if
    if (size(binEdges) .lt. hsize + 1) then
      status = -2
  return
    else if (size(binEdges) .gt. hsize + 1) then
      status = status + 2
    end if
    if (size(binIsNull) .lt. hsize) then
      status = -3
  return
    else if (size(binIsNull) .gt. hsize) then
      status = status + 4
    end if
    someNegativeErrors = .false. ! default
    do i = 1, hsize
      if (binErrors(i) .lt. 0.0) then
        someNegativeErrors = .true.
    exit
      end if
    end do
    if (someNegativeErrors) then
      status = -4
  return
    end if

    allocate(myBinEdges(hsize + 1), myBinContents(hsize), myBinErrors(hsize))
    myBinEdges = binEdges(1:hsize + 1)

    deviceID = pgopen(device)
    call pgsvp(0.07, 0.999, 0.1, 0.90)

    ! Amalgamate all bins which are smaller than a couple of pixels wide:
    call pgqvsz(3, leftEdgePixels, riteEdgePixels, botmEdgePixels&
    , toppEdgePixels)
    minBinWidth = (myBinEdges(hsize + 1) - myBinEdges(1)) * 3.0&
    / (riteEdgePixels - leftEdgePixels)
    myBinContents = 0.0 ! vector op, default
    myBinErrors   = 0.0 ! vector op, default
    newBin = 1
    someBinsWereAmalgamated = .false. ! default
    do oldBin = 1, hsize
      myBinContents(newBin) = myBinContents(newBin) + binContents(oldBin)
      myBinErrors(newBin)   = myBinErrors(newBin)   + binErrors(oldBin)
      if (myBinEdges(oldBin + 1) - myBinEdges(newBin) .lt. minBinWidth) then
        lastBinWasSmall = .true.
        someBinsWereAmalgamated = .true.

      else
        lastBinWasSmall = .false.
        myBinEdges(newBin + 1) = myBinEdges(oldBin + 1)
        newBin = newBin + 1
      end if
    end do

    if (lastBinWasSmall) then
      if (newBin .gt. 1) then ! amalgamate last bin (which is too thin to
      ! be left alone) with the last-but-one new bin.
        myBinEdges(newBin) = myBinEdges(hsize + 1)
        myBinContents(newBin - 1) = myBinContents(newBin - 1)&
        + myBinContents(newBin)
        myBinErrors(newBin - 1)   = myBinErrors(newBin - 1)&
        + myBinErrors(newBin)

        newNumBins = newBin - 1

      else ! width of all the bins was less than a couple of pixels? Should
      ! not ever happen.
        myBinEdges(newBin + 1) = myBinEdges(hsize + 1)
        newNumBins = newBin
      end if
    else
      newNumBins = newBin - 1
    end if

    if (divideByBinWidth) then
      do bin = 1, newNumBins
        myBinContents(bin) = myBinContents(bin) / (myBinEdges(bin + 1)&
        - myBinEdges(bin))
      end do
    end if

    if (someBinsWereAmalgamated) then
      status = status + 8
    end if

    minx = myBinEdges(1)
    maxx = myBinEdges(newNumBins + 1)
    miny = min(0.0, minval(myBinContents(1:newNumBins)))
    maxy = max(0.0, maxval(myBinContents(1:newNumBins)))
    xEdgeWidth = 0.05 * (maxx - minx)
    yEdgeWidth = 0.05 * (maxy - miny)

    call pgswin(minx - xEdgeWidth, maxx + xEdgeWidth, miny - yEdgeWidth&
    , maxy + yEdgeWidth)

    call pgsci(2)
    call pgsfs(2)
!***** amend to cater for errors.
    do bin = 1, newNumBins
      call pgrect(myBinEdges(bin), myBinEdges(bin + 1), 0.0&
      , myBinContents(bin))
    end do
    call pgsci(1)
    call pgbox('BCNT', 0., 0, 'BCNT', 0., 0)
    call pglab(trim(xLabel), trim(yLabel), trim(label))
    call pgmove(0.0, miny - yEdgeWidth)
    call pgdraw(0.0, maxy + yEdgeWidth)
    call pgmove(minx - xEdgeWidth, 0.0)
    call pgdraw(maxx + xEdgeWidth, 0.0)
    call pgclos

  end function plotHistogramKernel

  !---------------------------------------------------------------------
  subroutine plotImageReal(image, nullValue, device, style, wait&
  , mouse, deviceId, status)
  !---------------------------------------------------------------------

    implicit none 

    real,         intent(in)            :: image(:,:)
    real,         intent(in),  optional :: nullValue
    character(*), intent(in),  optional :: device, style
    logical,      intent(in),  optional :: wait, mouse
    integer,      intent(out), optional :: deviceId, status

    character(255) :: localDevice, localStyle
    integer :: xsize, ysize, localDevId, localStatus
    logical :: localWait, withNulls, localMouse
    real :: localNullValue

    !-------------------------------------------------------------------
    ! Start:

!print *, 'plotImageReal'

    if (present(nullValue)) then
      localNullValue = real(nullValue)
      withNulls = .true.
    else 
      localNullValue = 0.0
      withNulls = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then
      localStyle = style
    else
      localStyle = 'bw'
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(mouse)) then
      localMouse = mouse
    else
      localMouse = .false.
    end if

    xsize = size(image, 1)
    ysize = size(image, 2)

    call plotImageKernel(real(image), xsize, ysize, trim(localDevice), trim(localStyle)&
    , localWait, localDevId, localStatus, withNulls, localNullValue, localMouse)

    if (present(deviceId))  deviceId  = localDevId
    if (present(status)) status = localStatus

  end subroutine plotImageReal

  !---------------------------------------------------------------------
  subroutine plotImageDouble(image, nullValue, device, style, wait&
  , mouse, deviceId, status)
  !---------------------------------------------------------------------

    implicit none 

    real(kind(0d0)), intent(in)            :: image(:,:)
    real(kind(0d0)), intent(in),  optional :: nullValue
    character(*),    intent(in),  optional :: device, style
    logical,         intent(in),  optional :: wait, mouse
    integer,         intent(out), optional :: deviceId, status

    character(255) :: localDevice, localStyle
    integer :: xsize, ysize, localDevId, localStatus
    logical :: localWait, withNulls, localMouse
    real :: localNullValue

    !-------------------------------------------------------------------
    ! Start:

    if (present(nullValue)) then
      localNullValue = real(nullValue)
      withNulls = .true.
    else 
      localNullValue = 0.0
      withNulls = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then
      localStyle = style
    else
      localStyle = 'bw'
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(mouse)) then
      localMouse = mouse
    else
      localMouse = .false.
    end if

    xsize = size(image, 1)
    ysize = size(image, 2)

    call plotImageKernel(real(image), xsize, ysize, trim(localDevice), trim(localStyle)&
    , localWait, localDevId, localStatus, withNulls, localNullValue, localMouse)

    if (present(deviceId))  deviceId  = localDevId
    if (present(status)) status = localStatus

  end subroutine plotImageDouble

  !---------------------------------------------------------------------
  subroutine plotImageInt(image, nullValue, device, style, wait&
  , mouse, deviceId, status)
  !---------------------------------------------------------------------

    implicit none 

    integer,      intent(in)            :: image(:,:)
    integer,      intent(in),  optional :: nullValue
    character(*), intent(in),  optional :: device, style
    logical,      intent(in),  optional :: wait, mouse
    integer,      intent(out), optional :: deviceId, status

    character(255) :: localDevice, localStyle
    integer :: xsize, ysize, localDevId, localStatus
    logical :: localWait, withNulls, localMouse
    real :: localNullValue

    !-------------------------------------------------------------------
    ! Start:

    if (present(nullValue)) then
      localNullValue = real(nullValue)
      withNulls = .true.
    else 
      localNullValue = 0.0
      withNulls = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then
      localStyle = style
    else
      localStyle = 'bw'
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    if (present(mouse)) then
      localMouse = mouse
    else
      localMouse = .false.
    end if

    xsize = size(image, 1)
    ysize = size(image, 2)

    call plotImageKernel(real(image), xsize, ysize, trim(localDevice), trim(localStyle)&
    , localWait, localDevId, localStatus, withNulls, localNullValue, localMouse)

    if (present(deviceId))  deviceId  = localDevId
    if (present(status)) status = localStatus

  end subroutine plotImageInt

  !---------------------------------------------------------------------
  subroutine plotImageLogical(image, nullValue, device, style, wait&
  , mouse, deviceId, status)
  !---------------------------------------------------------------------

    implicit none 

    logical, intent(in) :: image(:,:)
    logical, intent(in), optional :: nullValue
    character(*), intent(in), optional :: device, style
    logical, intent(in), optional :: wait, mouse
    integer, intent(out), optional :: deviceId, status

    character(255) :: localDevice, localStyle
    integer :: xsize, ysize, localDevId, localStatus
    real, allocatable :: myImage(:,:)
    logical :: localWait, withNulls, localMouse
    real :: localNullValue

    !-------------------------------------------------------------------
    ! Start:

    if (present(nullValue)) then
      if (nullValue) then
        localNullValue = 1.0
      else
        localNullValue = 0.0
      end if
      withNulls = .true.
    else 
      localNullValue = 0.0
      withNulls = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(style)) then
      localStyle = style
    else
      localStyle = 'bw'
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    localMouse = .false. ! doesn't make sense for logical.

    xsize = size(image, 1)
    ysize = size(image, 2)
    allocate(myImage(xsize, ysize))
    where(image)
      myImage = 1.0
    elsewhere
      myImage = 0.0
    end where

    call plotImageKernel(myImage, xsize, ysize, trim(localDevice), trim(localStyle)&
    , localWait, localDevId, localStatus, withNulls, localNullValue, localMouse)

    if (present(deviceId))  deviceId  = localDevId
    if (present(status)) status = localStatus

  end subroutine plotImageLogical

  !---------------------------------------------------------------------
  subroutine plotImageKernel(image, xsize, ysize, device, style, wait&
  , deviceId, status, withNulls, nullValue, mouse)
  !---------------------------------------------------------------------

    implicit none 

    integer, intent(in) :: xsize, ysize
    real, intent(in) :: image(xsize, ysize), nullValue
    character(*), intent(in) :: device, style
    logical, intent(in) :: wait, withNulls, mouse
    integer, intent(inout) :: deviceId
    integer, intent(out) :: status

    real :: minValue, maxValue, csrX, csrY
    integer :: pgopen, deviceStatus, csrStatus, pgcurs, xi, yi
    logical :: openNewDevice
    character(1) :: csrChar

    !-------------------------------------------------------------------
    ! Start:
!*** null pixels not supported yet.
!print *, 'plotImageKernel', '*'//style//'*', len(style), '*'//device//'*'

    status = 0 ! default

    call initTrMatrix

deviceStatus = queryDeviceStatus(deviceId)
if (deviceStatus .le. 1) then
  deviceId = pgopen(trim(device))
else if (deviceStatus .eq. 2) then
  call pgslct(deviceId)
end if

    call pgsvp(0.0, 0.999, 0.0, 0.999)
    call pgswin(0.5, real(xsize) + 0.5, 0.5, real(ysize) + 0.5)

    minValue = minval(image)
    maxValue = maxval(image)

    select case(style)
    case('bw')
      call pggray(image, xsize, ysize, 1, xsize, 1, ysize, maxValue&
      , minValue, trMatrix)
!print *, 'ex pggray'

    case('colour', 'color')
      call findClrLimits(status)
      if (status .lt. 0) then
        call pgclos
  return
      end if

      call pgscir(minClrIndex, maxClrIndex)
      if (.not. coloursAlreadyLoaded) call loadDefaultColourMap
      call setDeviceColourMap(status)
      if (status .lt. 0) then
        call pgclos
  return
      end if

      call pgimag(image, xsize, ysize, 1, xsize, 1, ysize, minValue&
      , maxValue, trMatrix)

    case default
      status = -1
      call pgclos
  return
    end select

    call pgsci(1)
    call pgsfs(2)
    call pgbox('BC', 0.0, 0, 'BC', 0.0, 0)

    if (mouse) then
      do
        csrStatus = pgcurs(csrX, csrY, csrChar)
      if (csrStatus .ne. 1) exit
      if (csrChar .ne. 'A' .and. csrChar .ne. 'D' .and. csrChar .ne. 'X') exit

        xi = min(xsize, max(1, nint(csrX)))
        yi = min(ysize, max(1, nint(csrY)))
        print *, xi, yi, image(xi, yi)
      end do
    end if

    if (.not. wait) call pgclos

  end subroutine plotImageKernel

  !---------------------------------------------------------------------
  subroutine drawCross(xmin, xmid, xmax, ymin, ymid, ymax)
  !---------------------------------------------------------------------
  ! Assumes pgplot is already set up.

    implicit none 

    real, intent(in) :: xmin, xmid, xmax, ymin, ymid, ymax

    !-------------------------------------------------------------------
    ! Start:

    call pgmove(xmin, ymid)
    call pgdraw(xmax, ymid)
    call pgmove(xmid, ymin)
    call pgdraw(xmid, ymax)

  end subroutine drawCross

  !---------------------------------------------------------------------
  subroutine drawXticks(xmin, xmax, ymin, ymax)
  !---------------------------------------------------------------------
  ! Assumes pgplot is already set up.

    implicit none 

    real, intent(in) :: xmin, xmax, ymin, ymax
    real, parameter ::&
      minorTickFracLength = 0.013,&
      majorTickFracLength = 0.02
    real :: x, textx, texty, minorTickLength, majorTickLength
    integer :: nmin, nmax, n, m
    character(len=20) :: exp

    !-------------------------------------------------------------------
    ! Start:

    minorTickLength = minorTickFracLength * (ymax - ymin)
    majorTickLength = majorTickFracLength * (ymax - ymin)

!    y = ymin
!y = 0.0
    nmin = floor(xmin)
    nmax = floor(xmax)
    do n = nmin, nmax
      x = real(n)
      if (x .gt. xmin .and. x .lt. xmax) then
        call pgmove(x, ymin)
        call pgdraw(x, ymin + majorTickLength)
        call pgmove(x, ymax)
        call pgdraw(x, ymax - majorTickLength)

        textx = n - (xmax - xmin) * 0.018
        texty = ymin - (ymax - ymin) * 0.041
        call pgsch(1.0)
        call pgtext(textx, texty, '10')
        textx = n + (xmax - xmin) * 0.0085
        texty = ymin - (ymax - ymin) * 0.025
        call pgsch(0.7)
!        write(exp, '(i3)') n
        call leftJustifyInteger(n, exp)
        call pgtext(textx, texty, exp)
      end if
      do m = 2, 9
        x = n + log(real(m)) / myLog10
        if (x .gt. xmin .and. x .lt. xmax) then
          call pgmove(x, ymin)
          call pgdraw(x, ymin + minorTickLength)
          call pgmove(x, ymax)
          call pgdraw(x, ymax - minorTickLength)
        end if
      end do
    end do
    call pgsch(1.0)

  end subroutine drawXticks

  !---------------------------------------------------------------------
  subroutine drawYticks(xmin, xmax, ymin, ymax)
  !---------------------------------------------------------------------
  ! Assumes pgplot is already set up.

    implicit none 

    real, intent(in) :: xmin, xmax, ymin, ymax
    real, parameter ::&
      minorTickFracLength = 0.00433,&
      majorTickFracLength = 0.00667
    real :: y, textx, texty, minorTickLength, majorTickLength
    integer :: nmin, nmax, n, m
    character(len=20) :: exp

    !-------------------------------------------------------------------
    ! Start:

    minorTickLength = minorTickFracLength * (ymax - ymin)
    majorTickLength = majorTickFracLength * (ymax - ymin)

!    y = ymin
!y = 0.0
    nmin = floor(ymin)
    nmax = floor(ymax)
    do n = nmin, nmax
      y = real(n)
      if (y .gt. ymin .and. y .lt. ymax) then
        call pgmove(xmin, y)
        call pgdraw(xmin + majorTickLength, y)
        call pgmove(xmax, y)
        call pgdraw(xmax - majorTickLength, y)

        textx = xmax + (xmax - xmin) * 0.01
        texty = n - (ymax - ymin) * 0.009
        call pgsch(1.0)
        call pgtext(textx, texty, '10')
        textx = xmax + (xmax - xmin) * 0.037
        texty = n + (ymax - ymin) * 0.007
        call pgsch(0.7)
!        write(exp, '(i3)') n
        call leftJustifyInteger(n, exp)
        call pgtext(textx, texty, exp)
      end if
      do m = 2, 9
        y = n + log(real(m)) / myLog10
        if (y .gt. ymin .and. y .lt. ymax) then
          call pgmove(xmin, y)
          call pgdraw(xmin + minorTickLength, y)
          call pgmove(xmax, y)
          call pgdraw(xmax - minorTickLength, y)
        end if
      end do
    end do
    call pgsch(1.0)

  end subroutine drawYticks

  !---------------------------------------------------------------------
  subroutine leftJustifyInteger(i, str)
  !---------------------------------------------------------------------
  ! Assumes pgplot is already set up.

    implicit none 

    integer, intent(in) :: i
    integer :: intLength
    character(len=20), intent(out) :: str
!    character(len=1) :: char

    !-------------------------------------------------------------------
    ! Start:

    if (i .gt. 0) then
      write(str, '(i20)') i
      intLength = 1 + floor(log(real(i)) / myLog10)
!print *, 'lll', i, real(i), log(real(i)), log(real(i)) / myLog10, intLength
      str = str(21-intLength:)
    else if (i .eq. 0) then
      str = '0'
    else
      write(str, '(i20)') i
      intLength = 1 + floor(log(real(-i)) / myLog10)
!print *, 'lll', i, real(i), log(real(i)), log(real(i)) / myLog10, intLength
      str = str(20-intLength:)
    end if

  end subroutine leftJustifyInteger

  !---------------------------------------------------------------------
  subroutine plotColour3d(array, minZ, maxZ, obsThetaDeg, obsPhiDeg&
  , withGrid, device, wait, deviceId, status)
  !---------------------------------------------------------------------

    implicit none

    real,         intent(in)            :: array(:,:)
    real,         intent(in),  optional :: minZ,&
                                           maxZ,&
                                           obsThetaDeg,&
                                           obsPhiDeg
    logical,      intent(in),  optional :: withGrid,&
                                           wait
    character(*), intent(in),  optional :: device
    integer,      intent(out), optional :: deviceId, status

    real ::&
      localMinZ, localMaxZ, realXsizeMinus1, realYsizeMinus1,&
      delZ, boxDelZ, oldRed, oldGrn, oldBlu, localObsThetaDeg
    real, allocatable :: arrayX(:), arrayY(:), arrayZ(:,:)
    logical :: localWait, localWithGrid
    character(255) :: localDevice
    integer :: xsize, ysize, xi, yi, zi, pgopen, oldCI, oldFS, octant,&
      startXi, finisXi, stepXi, startYi, finisYi, stepYi, localDevId,&
      localStatus
    type(obsType) :: myObs
    type(pointType) :: box(2, 2, 2) ! the bounding box of the displayed surface.
    type(pointType) :: square(2, 2)

    !-------------------------------------------------------------------
    ! Start

    if (present(status)) status = 0 ! default

    call initDefaultObs

    xsize = size(array, 1)
    ysize = size(array, 2)
    allocate(arrayX(xsize), arrayY(ysize), arrayZ(xsize, ysize))

    if (present(minZ)) then
      localMinZ = minZ
    else
      localMinZ = minval(array)
    end if
    if (present(maxZ)) then
      localMaxZ = maxZ
    else
      localMaxZ = maxval(array)
    end if

    myObs%locRadius = defaultObs%locRadius

    if (present(obsThetaDeg)) then
      localObsThetaDeg = obsThetaDeg
      do while (localObsThetaDeg .lt. 0.0)
        localObsThetaDeg = localObsThetaDeg + 360.0
      end do
      do while (localObsThetaDeg .ge. 360.0)
        localObsThetaDeg = localObsThetaDeg - 360.0
      end do
      myObs%locTheta = localObsThetaDeg * myPiOn180
    else
      myObs%locTheta = defaultObs%locTheta
    end if

    if (present(obsPhiDeg)) then
      myObs%locPhi = obsPhiDeg * myPiOn180
    else
      myObs%locPhi = defaultObs%locPhi
    end if
    if (myObs%locPhi .lt. -myPiOn2 .or. myObs%locPhi .gt. myPiOn2) then
      if (present(status)) status = -1
  return
    end if

    if (present(withGrid)) then
      localWithGrid = withGrid
    else
      localWithGrid = .false.
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(wait)) then
      localWait = wait
    else
      localWait = .false.
    end if

    box(1, 1:2, 1:2)%solid(1) = 0.0
    box(2, 1:2, 1:2)%solid(1) = 1.0
    box(1:2, 1, 1:2)%solid(2) = 0.0
    box(1:2, 2, 1:2)%solid(2) = 1.0
    box(1:2, 1:2, 1)%solid(3) = 0.0
    box(1:2, 1:2, 2)%solid(3) = 0.5

    call completeObs(myObs, box)

    boxDelZ = box(1, 1, 2)%solid(3) - box(1, 1, 1)%solid(3)
    delZ = localMaxZ - localMinZ
    arrayZ = ((array - localMinZ) / delZ) * boxDelZ
    where (arrayZ .gt. boxDelZ) arrayZ = boxDelZ
    where (arrayZ .lt. 0.0) arrayZ = 0.0

    realXsizeMinus1 = real(xsize - 1)
    realYsizeMinus1 = real(ysize - 1)
    do xi = 1, xsize
      arrayX(xi) = (xi - 1) / realXsizeMinus1
    end do
    do yi = 1, ysize
      arrayY(yi) = (yi - 1) / realYsizeMinus1
    end do

!deviceStatus = queryDeviceStatus(deviceId)
!if (deviceStatus .le. 1)
!  deviceId = pgopen(trim(device))
!else if (deviceStatus .eq. 2)
!  call pgslct(deviceId)
!end if
    localDevId = pgopen(trim(localDevice))
    if (present(deviceId)) deviceId = localDevId
    call findClrLimits(localStatus)
    if (localStatus .lt. 0) then
      if (present(status)) status = localStatus
 return
    end if

    call pgsvp(0.0, 0.999, 0.0, 0.999)
    call pgswin(-0.5, 0.5, -0.5, 0.5)
    if (.not. coloursAlreadyLoaded) call loadDefaultColourMap
    call setDeviceColourMap(localStatus)
    if (localStatus .lt. 0) then
      if (present(status)) status = localStatus
 return
    end if

    call pgqfs(oldFS)
    call pgqci(oldCI)
    call pgqcr(2, oldRed, oldGrn, oldBlu)
    call pgscr(2, 0.5, 0.5, 0.5) ! grey
    call pgsfs(1) ! filled

    ! Calc octant:
    octant = 1 + floor(8.0 * (myObs%locTheta / myTwoPi))

    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          box(xi, yi, zi)%proj = project(box(xi, yi, zi)%solid, myObs)
        end do
      end do
    end do

    call pgsci(1)

    select case(octant)
    case(1, 2)
      startXi = 1
      finisXi = xsize - 1
      stepXi = 1
      startYi = 1
      finisYi = ysize - 1
      stepYi = 1

      ! Draw all axes:
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))

    case(3, 4)
      startXi = xsize - 1
      finisXi = 1
      stepXi = -1
      startYi = 1
      finisYi = ysize - 1
      stepYi = 1

      ! Draw x axis:
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))

    case(5, 6)
      startXi = xsize - 1
      finisXi = 1
      stepXi = -1
      startYi = ysize - 1
      finisYi = 1
      stepYi = -1

      ! Draw no axes.

    case(7, 8)
      startXi = 1
      finisXi = xsize - 1
      stepXi = 1
      startYi = ysize - 1
      finisYi = 1
      stepYi = -1

      ! Draw y axis:
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))

    case default
      if (present(status)) status = -3
  return
    end select

    do xi = startXi, finisXi, stepXi
!    if (xi .ne. 11) cycle
      do yi = startYi, finisYi, stepYi
!      if (yi .ne. 47) cycle
        square(1,1)%solid = (/arrayX(xi),   arrayY(yi),   arrayZ(xi, yi)/)
        square(2,1)%solid = (/arrayX(xi+1), arrayY(yi),   arrayZ(xi+1, yi)/)
        square(1,2)%solid = (/arrayX(xi),   arrayY(yi+1), arrayZ(xi, yi+1)/)
        square(2,2)%solid = (/arrayX(xi+1), arrayY(yi+1), arrayZ(xi+1, yi+1)/)
        call processSquare(square, boxDelZ, myObs, octant, localWithGrid&
        , localStatus)
        if (localStatus .lt. 0) then
          if (present(status)) status = localStatus
 return
        end if
      end do
    end do

    ! Draw the rest of the axes:
    call pgsci(1)
    select case(octant)
    case(1, 2) ! Draw no axes.

    case(3, 4) ! Draw y, z axes:
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))

    case(5, 6) ! Draw all axes:
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))

    case(7, 8) ! Draw x, z axes:
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))

    case default
      if (present(status)) status = -3
  return
    end select

    call pgbox('BC', 0.0, 0, 'BC', 0.0, 0)

!    call pgsfs(2) ! hollow
!    call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
!    call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
!    call pgdraw(box(2, 2, 1)%proj(1), box(2, 2, 1)%proj(2))
!    call pgdraw(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
!    call pgdraw(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
!    call pgmove(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))
!    call pgdraw(box(1, 2, 2)%proj(1), box(1, 2, 2)%proj(2))
!    call pgdraw(box(2, 2, 2)%proj(1), box(2, 2, 2)%proj(2))
!    call pgdraw(box(2, 1, 2)%proj(1), box(2, 1, 2)%proj(2))
!    call pgdraw(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))

!    call pgscr(2, oldRed, oldGrn, oldBlu)
    call pgsfs(oldFS)
    call pgsci(oldCI)

    if (.not. localWait) call pgclos

  end subroutine plotColour3d

  !---------------------------------------------------------------------
  subroutine loadDefaultColourMap
  !---------------------------------------------------------------------

    implicit none

    integer :: clrIndex
    real :: frac

    !-------------------------------------------------------------------
    ! Start

    do clrIndex = numReservedClrs, 255
      frac = (clrIndex - numReservedClrs) / real(255 - numReservedClrs)
      colour(clrIndex)%hue(red) = frac
      colour(clrIndex)%hue(grn) = frac
      colour(clrIndex)%hue(blu) = 0.3 * (1.0 - frac)
    end do

    coloursAlreadyLoaded = .true.

  end subroutine loadDefaultColourMap

  !---------------------------------------------------------------------
  subroutine setDeviceColourMap(status)
  !---------------------------------------------------------------------
!**** turn it into a function returning status.

    implicit none

    integer, intent(out) :: status

    integer :: clrIndex, hueIndex
    real :: frac

    !-------------------------------------------------------------------
    ! Start

    status = 0

    do clrIndex = minClrIndex, maxClrIndex + 1
      frac = (clrIndex - minClrIndex) / real(numClrIndices)
      newXedges(clrIndex) = frac
    end do
    
    do hueIndex = red, blu
      call rebin(oldXedges(numReservedClrs:256)&
      , colour(numReservedClrs:255)%hue(hueIndex)&
      , newXedges(minClrIndex:maxClrIndex + 1)&
      , deviceColour(minClrIndex:maxClrIndex)%hue(hueIndex)&
      , status=status)
  if (status .lt. 0) return

      deviceColour%hue(hueIndex) = deviceColour%hue(hueIndex)&
      * real(numClrIndices) / real(256 - numReservedClrs)

      where(deviceColour%hue(hueIndex) .gt. 1.0)&
        deviceColour%hue(hueIndex) = 1.0
      where(deviceColour%hue(hueIndex) .lt. 0.0)&
        deviceColour%hue(hueIndex) = 0.0
    end do

    do clrIndex = minClrIndex, maxClrIndex
      call pgscr(clrIndex&
      , deviceColour(clrIndex)%hue(red)&
      , deviceColour(clrIndex)%hue(grn)&
      , deviceColour(clrIndex)%hue(blu))
    end do

  end subroutine setDeviceColourMap

  !---------------------------------------------------------------------
  subroutine processSquare(square, delZ, obs, octant, withGrid, status)
  !---------------------------------------------------------------------

    implicit none

    type(pointType), intent(in) :: square(2, 2)
    type(obsType), intent(in) :: obs
    integer, intent(in) :: octant
    real, intent(in) :: delZ
    logical, intent(in) :: withGrid
    integer, intent(out) :: status

    type(pointType) :: triangle(4, 3)
    integer :: order(4), i
    real :: corner1(2), corner2(2)

    !-------------------------------------------------------------------
    ! Start

    status = 0

    triangle(1:4, 3)%solid(1) = (square(1, 1)%solid(1) + square(2, 1)%solid(1))&
    / 2.0
    triangle(1:4, 3)%solid(2) = (square(1, 1)%solid(2) + square(1, 2)%solid(2))&
    / 2.0
    triangle(1:4, 3)%solid(3) = (square(1, 1)%solid(3) + square(1, 2)%solid(3)&
    + square(2, 1)%solid(3) + square(2, 2)%solid(3)) / 4.0

    triangle(1, 1) = square(1, 1)
    triangle(1, 2) = square(2, 1)
    triangle(2, 1) = square(2, 1)
    triangle(2, 2) = square(2, 2)
    triangle(3, 1) = square(2, 2)
    triangle(3, 2) = square(1, 2)
    triangle(4, 1) = square(1, 2)
    triangle(4, 2) = square(1, 1)

    select case(octant)
    case(8, 1)
      order = (/4, 1, 3, 2/)
    case(2, 3)
      order = (/1, 4, 2, 3/)
    case(4, 5)
      order = (/2, 3, 1, 4/)
    case(6, 7)
      order = (/3, 2, 4, 1/)
    case default
      status = -3
  return
    end select

!do i = 7, maxClrIndex
!call pgscr(i, 0.0, 0.0, 0.0)
!end do
!call pgscr(3, 1.0, 0.0, 0.0)
!call pgscr(4, 1.0, 1.0, 0.0)
!call pgscr(5, 0.0, 1.0, 0.0)
!call pgscr(6, 0.0, 0.0, 1.0)

    call pgsfs(1) ! filled
    do i = 1, 4
!    if (i .ne. 3) cycle
      call processTriangle(triangle(order(i), 1:3), delZ, obs)

      if (withGrid) then
        call pgsci(2)
        corner1 = project(triangle(order(i), 1)%solid, obs)
        corner2 = project(triangle(order(i), 2)%solid, obs)
        call pgmove(corner1(1), corner1(2))
        call pgdraw(corner2(1), corner2(2))
      end if
    end do

  end subroutine processSquare

  !---------------------------------------------------------------------
  subroutine processTriangle(triangle, delZ, obs)
  !---------------------------------------------------------------------
  ! At this point we have 3 points (necessarily defining a triangle) in
  ! 3d space. We want to project the coordinates of its corners onto the
  ! image plane and then draw its (filled) shape on this plane. However
  ! the triangle needs to be coloured in different zones, the points in
  ! each zone having z values within a certain range. Therefore before
  ! performing the projection/plotting step it is necessary to break the
  ! triangle into these zones and treat each one separately. Another way
  ! to think of this is to define a number of parallel planes within the
  ! 3d space occupied by the triangle, each of constant z, and spaced
  ! delZ/numClrIndices apart. The triangular surface must be broken into
  ! polygons, each representing that part of the triangle which is
  ! contained between two neighbouring planes.
  !
  ! The height (ie, delta z) of that side of the triangle with the largest
  ! height (call it the long-z side) is always equal to the sum of the
  ! other two delta zs (I'll call these sides 'top' and 'bot'). It is
  ! simpler to start each polygon (which can't have more than 5 sides)
  ! on the long-z side because, starting from the bottom, the edge of the
  ! polygon along this side of the triangle never encompasses a vertex, but
  ! at 1 and only 1 point on the other side, there will be a vertex. Thus
  ! we know where we are. Points 1 and 2 of the polygon are therefore
  ! always defined as the top side of the poly, point 1 being on the
  ! 'long-z' side of the triangle; this also allows them to be re-used as
  ! the appropriate bottom points of the next higher polygon.

    implicit none

    type(pointType), intent(in) :: triangle(3)
    type(obsType), intent(in) :: obs
    real, intent(in) :: delZ
    type(pointType) :: poly(5)
    integer :: iOfMinZ, iOfMedZ, iOfMaxZ, startCI, finisCI, n, numPolyPoints, clrIndex
    real :: plotX(5), plotY(5), z
    logical :: passedVertex

    !-------------------------------------------------------------------
    ! Start

!*** use of a sort routine would be better.

    if (triangle(1)%solid(3) .gt. triangle(2)%solid(3)) then
      if (triangle(1)%solid(3) .gt. triangle(3)%solid(3)) then
        if (triangle(2)%solid(3) .gt. triangle(3)%solid(3)) then
!print *, '1 > 2 > 3'
          iOfMaxZ = 1
          iOfMedZ = 2
          iOfMinZ = 3
        else
!print *, '1 > 3 >= 2'
          iOfMaxZ = 1
          iOfMedZ = 3
          iOfMinZ = 2
        end if
      else
!print *, '3 >= 1 > 2'
        iOfMaxZ = 3
        iOfMedZ = 1
        iOfMinZ = 2
      end if
    else
      if (triangle(2)%solid(3) .gt. triangle(3)%solid(3)) then
        if (triangle(1)%solid(3) .gt. triangle(3)%solid(3)) then
!print *, '2 >= 1 > 3'
          iOfMaxZ = 2
          iOfMedZ = 1
          iOfMinZ = 3
        else
!print *, '2 >= 3 >= 1'
          iOfMaxZ = 2
          iOfMedZ = 3
          iOfMinZ = 1
        end if
      else
!print *, '3 >= 2 >= 1'
        iOfMaxZ = 3
        iOfMedZ = 2
        iOfMinZ = 1
      end if
    end if
!print *, ' '
!print *, 'minClrIndex, delZ, numClrIndices:', minClrIndex, delZ, numClrIndices

    startCI = minClrIndex + ceiling(triangle(iOfMinZ)%solid(3) / delZ&
    * numClrIndices)
!    startCI = min(max(startCI, minClrIndex + 1), maxClrIndex)
startCI = min(max(startCI, minClrIndex + 1), maxClrIndex + 1)
    finisCI = minClrIndex + floor(triangle(iOfMaxZ)%solid(3) / delZ&
    * numClrIndices)
    finisCI = min(max(finisCI, minClrIndex), maxClrIndex)

!print *, triangle(iOfMinZ)%solid(3), startCI
!print *, triangle(iOfMedZ)%solid(3)
!print *, triangle(iOfMaxZ)%solid(3), finisCI
!print *, ' '

    if (startCI .gt. finisCI) then ! triangle lies totally within a single
!if (startCI .ge. finisCI) then ! triangle lies totally within a single
    ! colour index stratum.
      poly(1)%proj = project(triangle(1)%solid, obs)
      poly(2)%proj = project(triangle(2)%solid, obs)
      poly(3)%proj = project(triangle(3)%solid, obs)
!print *, 'single colour.'
!print *, poly(1)%proj(1), poly(1)%proj(2)
!print *, poly(2)%proj(1), poly(2)%proj(2)
!print *, poly(3)%proj(1), poly(3)%proj(2)
!print *, ' '

      plotX(1:3) = (/poly(1)%proj(1), poly(2)%proj(1), poly(3)%proj(1)/)
      plotY(1:3) = (/poly(1)%proj(2), poly(2)%proj(2), poly(3)%proj(2)/)
      call pgsci(finisCI)
      call pgpoly(3, plotX, plotY)
  return
    end if

!print *, 'many colours.'

    clrIndex = startCI
    z = delZ * (clrIndex - minClrIndex) / real(numClrIndices)
    call temp_interpolate(triangle(iOfMinZ), triangle(iOfMaxZ), z, poly(1))
!print *, 'ci, z:', clrIndex, z

    passedVertex = .false.
    if (z .gt. triangle(iOfMedZ)%solid(3)) then ! include the vertex
      call temp_interpolate(triangle(iOfMedZ), triangle(iOfMaxZ), z, poly(2))
      poly(3) = triangle(iOfMedZ)
      poly(4) = triangle(iOfMinZ)
      numPolyPoints = 4
      passedVertex = .true.

    else
      call temp_interpolate(triangle(iOfMinZ), triangle(iOfMedZ), z, poly(2))
      poly(3) = triangle(iOfMinZ)
      numPolyPoints = 3
    end if
    do n = 1, numPolyPoints
      poly(n)%proj = project(poly(n)%solid, obs)
      plotX(n) = poly(n)%proj(1)
      plotY(n) = poly(n)%proj(2)
    end do
    call pgsci(startCI - 1)
    call pgpoly(numPolyPoints, plotX, plotY)

    do clrIndex = startCI + 1, finisCI
      z = delZ * (clrIndex - minClrIndex) / real(numClrIndices)
!print *, 'ci, z:', clrIndex, z

      if (passedVertex) then
        poly(3) = poly(2) ! do these out of order otherwise poly(1) and (2)
        poly(4) = poly(1) ! are already overwritten, dumbo.

        call temp_interpolate(triangle(iOfMedZ), triangle(iOfMaxZ), z, poly(2))
        numPolyPoints = 4

      else if (z .gt. triangle(iOfMedZ)%solid(3)) then
      ! include the vertex
        poly(4) = poly(2) ! do these out of order otherwise poly(1) and (2)
        poly(5) = poly(1) ! are already overwritten, dumbo.

        call temp_interpolate(triangle(iOfMedZ), triangle(iOfMaxZ), z, poly(2))
        poly(3) = triangle(iOfMedZ)
        numPolyPoints = 5
        passedVertex = .true.

      else
        poly(3) = poly(2) ! do these out of order otherwise poly(1) and (2)
        poly(4) = poly(1) ! are already overwritten, dumbo.

        call temp_interpolate(triangle(iOfMinZ), triangle(iOfMedZ), z, poly(2))
        numPolyPoints = 4
      end if
      call temp_interpolate(triangle(iOfMinZ), triangle(iOfMaxZ), z, poly(1))

      do n = 1, numPolyPoints
        poly(n)%proj = project(poly(n)%solid, obs)
        plotX(n) = poly(n)%proj(1)
        plotY(n) = poly(n)%proj(2)
      end do
      call pgsci(clrIndex - 1)
      call pgpoly(numPolyPoints, plotX, plotY)

    end do

    if (passedVertex) then
!     poly(2) = poly(2) ! do these out of order otherwise poly(1) and (2)
      poly(3) = poly(1) ! are already overwritten, dumbo.

      numPolyPoints = 3

    else
      poly(3) = poly(2) ! do these out of order otherwise poly(1) and (2)
      poly(4) = poly(1) ! are already overwritten, dumbo.

      poly(2) = triangle(iOfMedZ)
      numPolyPoints = 4
    end if
    poly(1) =  triangle(iOfMaxZ)

    do n = 1, numPolyPoints
      poly(n)%proj = project(poly(n)%solid, obs)
      plotX(n) = poly(n)%proj(1)
      plotY(n) = poly(n)%proj(2)
    end do
    call pgsci(finisCI)
    call pgpoly(numPolyPoints, plotX, plotY)

  end subroutine processTriangle

  !---------------------------------------------------------------------
  subroutine temp_interpolate(loPoint, hiPoint, z, zPoint)
  !---------------------------------------------------------------------
!*** use the interpolate routine in mathutils?

    implicit none

    type(pointType), intent(in) :: loPoint, hiPoint
    real, intent(in) :: z
    type(pointType), intent(out) :: zPoint
    real :: f

    !-------------------------------------------------------------------
    ! Start

    zPoint%solid(3) = z
    f = (z - loPoint%solid(3)) / (hiPoint%solid(3) - loPoint%solid(3))
    zPoint%solid(1:2) = loPoint%solid(1:2) * (1.0 - f) + hiPoint%solid(1:2) * f

  end subroutine temp_interpolate

  !---------------------------------------------------------------------
  subroutine completeObs(obs, box)
  !---------------------------------------------------------------------

    implicit none

    type(obsType), intent(inout) :: obs
! has real :: location(3), locRadius, locTheta, locPhi, viewFrame(3, 3), viewTheta, viewPhi, viewTwist, projOffsets(2), projScale
    type(pointType), intent(inout) :: box(2, 2, 2)
    real :: obsSemiradius, boxCentre(3), minBoxProj(2), maxBoxProj(2)
    integer :: xi, yi, zi

    !-------------------------------------------------------------------
    ! Start

    obsSemiradius = obs%locRadius * cos(obs%locPhi)
    obs%location(1) = obsSemiradius * cos(obs%locTheta)
    obs%location(2) = obsSemiradius * sin(obs%locTheta)
    obs%location(3) = obs%locRadius * sin(obs%locPhi)

    boxCentre = 0.0 ! vector op
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          boxCentre = boxCentre + box(xi, yi, zi)%solid
        end do
      end do
    end do
    boxCentre = boxCentre / 8.0

    obs%viewFrame(3, 1:3) = normalise(obs%location - boxCentre) ! z axis of
    ! the viewer frame points from the scene to the viewer's eye.

    if (obs%viewFrame(3, 1) .eq. 0.0 .and. obs%viewFrame(3, 2) .eq. 0.0) then
    ! the direction of view is vertical. Choose x and y axes so the x axis
    ! lines up with the box basis x axis.
      obs%viewFrame(1, 1:3) = (/1.0, 0.0, 0.0/)

    else
      obs%viewFrame(1, 1) = -obs%viewFrame(3, 2)
      obs%viewFrame(1, 2) = obs%viewFrame(3, 1)
      obs%viewFrame(1, 3) = 0.0
      obs%viewFrame(1, 1:3) = normalise(obs%viewFrame(1, 1:3))
    end if

    obs%viewFrame(2, 1:3) = vectorCross(obs%viewFrame(3, 1:3)&
    , obs%viewFrame(1, 1:3))

    obs%viewFrame(2, 1:3) = normalise(obs%viewFrame(2, 1:3))

    obs%projOffsets = 0.0
    obs%projScale = 1.0
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          box(xi, yi, zi)%proj = project(box(xi, yi, zi)%solid, obs)
        end do
      end do
    end do

if (testFlag) then
print *, ' '
print *, 'completeObs: box centre:'
print *, boxCentre(1), boxCentre(2), boxCentre(3)
end if

    minBoxProj = box(1, 1, 1)%proj
    maxBoxProj = box(1, 1, 1)%proj
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          minBoxProj(1) = min(minBoxProj(1), box(xi, yi, zi)%proj(1))
          minBoxProj(2) = min(minBoxProj(2), box(xi, yi, zi)%proj(2))
          maxBoxProj(1) = max(maxBoxProj(1), box(xi, yi, zi)%proj(1))
          maxBoxProj(2) = max(maxBoxProj(2), box(xi, yi, zi)%proj(2))
        end do
      end do
    end do

    obs%projOffsets(1) = -(minBoxProj(1) + maxBoxProj(1)) / 2.0
    obs%projOffsets(2) = -(minBoxProj(2) + maxBoxProj(2)) / 2.0
    obs%projScale = 0.9 / max(maxBoxProj(1) - minBoxProj(1)&
    , maxBoxProj(2) - minBoxProj(2))

  end subroutine completeObs

  !---------------------------------------------------------------------
  subroutine fillOutObs(obs)
  !---------------------------------------------------------------------
  ! This sub calculates nearly all the other entries of the obs data
  ! structure given locPhi, locTheta and locRadius. The exceptions are
  ! projOffsets and projScale. 

    implicit none

    type(obsType), intent(inout) :: obs
! has real :: location(3), locRadius, locTheta, locPhi, viewFrame(3, 3), viewTheta, viewPhi, viewTwist, projOffsets(2), projScale
    real :: obsSemiradius

    !-------------------------------------------------------------------
    ! Start

    obsSemiradius = obs%locRadius * cos(obs%locPhi)
    obs%location(1) = obsSemiradius * cos(obs%locTheta)
    obs%location(2) = obsSemiradius * sin(obs%locTheta)
    obs%location(3) = obs%locRadius * sin(obs%locPhi)

    obs%viewFrame(3, 1:3) = normalise(obs%location) ! z axis of
    ! the viewer frame points from the scene to the viewer's eye.

    if (obs%viewFrame(3, 1) .eq. 0.0 .and. obs%viewFrame(3, 2) .eq. 0.0) then
    ! the direction of view is vertical. Choose x and y axes so the x axis
    ! lines up with the box basis x axis.
      obs%viewFrame(1, 1:3) = (/1.0, 0.0, 0.0/)

    else ! x axis of the view frame lies in the x-y plane of the lab frame, 90 degrees anticlockwise from the z axis of the view frame, viewed from above in the lab frame.
      obs%viewFrame(1, 1) = -obs%viewFrame(3, 2)
      obs%viewFrame(1, 2) = obs%viewFrame(3, 1)
      obs%viewFrame(1, 3) = 0.0
      obs%viewFrame(1, 1:3) = normalise(obs%viewFrame(1, 1:3))
    end if

    obs%viewFrame(2, 1:3) = vectorCross(obs%viewFrame(3, 1:3)&
    , obs%viewFrame(1, 1:3))

    obs%viewFrame(2, 1:3) = normalise(obs%viewFrame(2, 1:3))

  end subroutine fillOutObs

  !---------------------------------------------------------------------
  function project(pointIn3d, obs) result(projectedPoint)
  !---------------------------------------------------------------------
!****** this should be changed to a subroutine, and to return a status instead of call error().

    implicit none

    type(obsType), intent(in) :: obs
    real, intent(in) :: pointIn3d(3)
    real :: projectedPoint(2)
    real :: view(3), viewInObsFrame(3)

    !-------------------------------------------------------------------
    ! Start

    view = pointIn3d - obs%location
if (testFlag) then
print *, ' '
print *, '  project: view vector in array frame:'
print *, view(1), view(2), view(3)
end if

    viewInObsFrame&
    = view(1) * obs%viewFrame(1:3, 1)&
    + view(2) * obs%viewFrame(1:3, 2)&
    + view(3) * obs%viewFrame(1:3, 3)
if (viewInObsFrame(3) .ge. 0.0) call error('', 'point is behind the viewer.')
if (testFlag) then
print *, '  project: view vector in observer frame:'
print *, viewInObsFrame(1), viewInObsFrame(2), viewInObsFrame(3)
end if

    projectedPoint(1) = obs%projScale * (atan2(viewInObsFrame(1)&
    , -viewInObsFrame(3)) + obs%projOffsets(1))
    projectedPoint(2) = obs%projScale * (atan2(viewInObsFrame(2)&
    , -viewInObsFrame(3)) + obs%projOffsets(2))

  end function project

  !---------------------------------------------------------------------
  subroutine project_new(pointIn3d, obs, projectedPoint, status)
  !---------------------------------------------------------------------

    implicit none

    type(obsType), intent(in) :: obs
    real, intent(in) :: pointIn3d(3)
    real, intent(out) :: projectedPoint(2)
    integer, intent(out) :: status

    real :: view(3), viewInObsFrame(3)

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    view = pointIn3d - obs%location

    viewInObsFrame&
    = view(1) * obs%viewFrame(1:3, 1)&
    + view(2) * obs%viewFrame(1:3, 2)&
    + view(3) * obs%viewFrame(1:3, 3)

    if (viewInObsFrame(3) .ge. 0.0) then
      status = -1
  return
    end if

    projectedPoint(1) = obs%projScale * (atan2(viewInObsFrame(1)&
    , -viewInObsFrame(3)) + obs%projOffsets(1))
    projectedPoint(2) = obs%projScale * (atan2(viewInObsFrame(2)&
    , -viewInObsFrame(3)) + obs%projOffsets(2))

  end subroutine project_new

  !---------------------------------------------------------------------
  subroutine closePgDevice(deviceId)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in), optional :: deviceId

    !-------------------------------------------------------------------
    ! Start

    if (present(deviceId)) then
      call pgslct(deviceId)
      call pgclos
    else ! close all
      call pgend
    end if

  end subroutine closePgDevice

  !---------------------------------------------------------------------
  function editColourMap(deviceId, device) result(status)
  !---------------------------------------------------------------------
! ***** should have an error if deviceId is not open?? Can check to see if another device is open via pgqinf. Better than nothing...

    implicit none

    integer,      intent(in),  optional :: deviceId
    character(*), intent(in),  optional :: device
    integer :: status

    integer :: deviceID2, pgopen, btnNum, answerLength, pgcurs, clrIndex
    real :: x, y, newX, newY, huesOfReservedClrs(numReservedClrs, red:blu)
    character(len=1) :: char
    character(255) :: answer

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    if (present(device)) then
      deviceID2 = pgopen(device)
    else
      deviceID2 = pgopen(cmapModDevice)
    end if

    call pgqinf('hardcopy', answer, answerLength)
    if (answerLength .eq. 3 .and. index(answer, 'YES') .gt. 0) then
      status = -4
      call pgclos
  return
    end if

    call findECMclrLimits(status)
  if (status .lt. 0) return

!deviceStatus = queryDeviceStatus(deviceId)
!if (deviceStatus .le. 1)
!  deviceId = pgopen(trim(device))
!else if (deviceStatus .eq. 2)
!  call pgslct(deviceId)
!end if
    if (.not. present(deviceId)) then
      call loadDefaultColourMap
    end if
    call deducePins(pins)
    call snapColoursToPins
    call setDeviceColourMap(status)
  if (status .lt. 0) return

    call pgscr(2, 0.5, 0.5, 0.5)
    if (present(deviceId)) then ! change its colour map too:
      call pgslct(deviceId)
      call findClrLimits(status)
  if (status .lt. 0) return

      call setDeviceColourMap(status)
  if (status .lt. 0) return

      do clrIndex = 1, numReservedClrs - 1
        call pgqcr(clrIndex&
        , huesOfReservedClrs(clrIndex, red)&
        , huesOfReservedClrs(clrIndex, grn)&
        , huesOfReservedClrs(clrIndex, blu)&
        )
      end do
      call pgslct(deviceID2)
      do clrIndex = 1, numReservedClrs - 1
        call pgscr(clrIndex&
        , huesOfReservedClrs(clrIndex, red)&
        , huesOfReservedClrs(clrIndex, grn)&
        , huesOfReservedClrs(clrIndex, blu)&
        )
      end do
    end if

    call pgsvp(0.0, 0.999, 0.0, 0.999)
    call pgswin(0.0, 1.0, 0.0, 1.0)

    call drawInitialPlot!(pins)

    do
      status = pgcurs(x, y, char)
      btnNum = readButton(x, y)

      select case (btnNum)
      case (1) ! just exit
    exit

      case (2) ! load defaults
        call loadDefaultColourMap
        call deducePins(pins)
        call snapColoursToPins
        call setDeviceColourMap(status)
  if (status .lt. 0) return

        if (present(deviceId)) then ! change its colour map too:
          call pgslct(deviceId)
          call findClrLimits(status)
  if (status .lt. 0) return

          call setDeviceColourMap(status)
  if (status .lt. 0) return

          call pgslct(deviceID2)
        end if
        call pgeras
        selectedCi = -1
        selectedHi = -1
        call drawInitialPlot

      case (3) ! delete selected point
        call deletePoint(present(deviceId), deviceID2, deviceId, status)
  if (status .lt. 0) return

      case (4) ! undo all
!        call pgeras
!        call drawInitialPlot!(pins)

      case (5) ! undo last
!        call undoLast

      case (6) ! save
!        call saveColourmap

      case default ! no button, drag the relevant pinning diagram:
        call dragPin(x, y, newX, newY, present(deviceId), deviceID2, deviceId, status)
  if (status .lt. 0) return

        x = newX ! stops the
        y = newY ! cursor jumping.
      end select
    end do

!    deallocate(rgbValue, rgbPinned, rgbPinValue)
    call pgclos

  end function editColourMap 

  !---------------------------------------------------------------------
  subroutine findECMclrLimits(status)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(out) :: status

    !-------------------------------------------------------------------
    ! Start

    status = 0

    call findClrLimits(status)
  if (status .lt. 0) return

    minClrIndex4ECM = minClrIndex
    maxClrIndex4ECM = maxClrIndex
    numClrIndices4ECM = numClrIndices
    realNumCI4ECM = realNumCI
    realNumCIminus14ECM = realNumCIminus1

  end subroutine findECMclrLimits 

  !---------------------------------------------------------------------
  subroutine findClrLimits(status)
  !---------------------------------------------------------------------
!**** change to function.

    implicit none

    integer, intent(out) :: status

    integer :: clrIndex
    real :: frac

    !-------------------------------------------------------------------
    ! Start

    status = 0

    call pgqcol(minClrIndex, maxClrIndex)
    minClrIndex = max(minClrIndex, numReservedClrs)
    if (minClrIndex .ge. maxClrIndex) then
      status = -2
  return
    end if
    numClrIndices = maxClrIndex - minClrIndex + 1
    realNumCI = real(maxClrIndex + 1 - minClrIndex)
    realNumCIminus1 = real(maxClrIndex - minClrIndex)

    do clrIndex = numReservedClrs, 255
      frac = (clrIndex - numReservedClrs) / real(256 - numReservedClrs)
      oldXedges(clrIndex) = frac
    end do
    oldXedges(256) = 1.0

  end subroutine findClrLimits 

  !---------------------------------------------------------------------
  subroutine deducePins(pins)
  !---------------------------------------------------------------------
!**** there are probably better ways to do this.

    implicit none

    type(pinType), intent(out) :: pins(0:255)
    integer :: hueIndex, clrIndex
    real :: bend
    real, parameter :: minBend = 1.0 / 256.0 / 10.0
!**** should figure in numReservedClrs.

    !-------------------------------------------------------------------
    ! Start

    do hueIndex = red, blu
      pins%isPinned(hueIndex) = .true. ! vector op, default.
      pins%hue(hueIndex) = colour%hue(hueIndex) ! vector op, default.
      do clrIndex = numReservedClrs + 1, 254
        bend = colour(clrIndex + 1)%hue(hueIndex)&
        + colour(clrIndex - 1)%hue(hueIndex)&
        - 2.0 * colour(clrIndex)%hue(hueIndex)

        if (abs(bend) .lt. minBend) then
          pins(clrIndex)%isPinned(hueIndex) = .false.
        end if
      end do
    end do

  end subroutine deducePins 

  !---------------------------------------------------------------------
  subroutine snapColoursToPins
  !---------------------------------------------------------------------

    implicit none

    integer :: hueIndex, clrIndex, ci2, clrIndexOfLastPin
    logical :: lastWasPinned
    real :: frac

    !-------------------------------------------------------------------
    ! Start

    colour(numReservedClrs)%hue = pins(numReservedClrs)%hue
    do hueIndex = red, blu
      clrIndexOfLastPin = numReservedClrs
      lastWasPinned = .true.
      do clrIndex = numReservedClrs + 1, 255
        if (pins(clrIndex)%isPinned(hueIndex)) then
          if (lastWasPinned) then
            colour(clrIndex)%hue(hueIndex) = pins(clrIndex)%hue(hueIndex)
          else
            do ci2 = clrIndexOfLastPin + 1, clrIndex
              frac = (ci2 - clrIndexOfLastPin)&
              / real(clrIndex - clrIndexOfLastPin)

              colour(ci2)%hue(hueIndex) = pins(clrIndexOfLastPin)%hue(hueIndex)&
              * (1.0 - frac) + pins(clrIndex)%hue(hueIndex) * frac
            end do
          end if
          clrIndexOfLastPin = clrIndex
          lastWasPinned = .true.
        else
          lastWasPinned = .false.
        end if
      end do
    end do

  end subroutine snapColoursToPins 

  !---------------------------------------------------------------------
  subroutine drawInitialPlot!(pins)
  !---------------------------------------------------------------------

    implicit none

!    type(pinType), intent(in) :: pins(0:255)

    real :: boxWidth, frac, btnHeight, textX, textY, yEdgeFracLo, yEdgeFracHi,&
      yEdgeLo, yEdgeHi
    integer :: hueIndex, btnNum, clrIndex

    !-------------------------------------------------------------------
    ! Start

    call pgsfs(1)
    do clrIndex = minClrIndex4ECM, maxClrIndex4ECM
      call pgsci(clrIndex)
      yEdgeFracLo = (clrIndex - minClrIndex4ECM) / realNumCI4ECM
      yEdgeFracHi = (clrIndex + 1 - minClrIndex4ECM) / realNumCI4ECM
      yEdgeLo = clrStripMinY * (1.0 - yEdgeFracLo) + clrStripMaxY * yEdgeFracLo
      yEdgeHi = clrStripMinY * (1.0 - yEdgeFracHi) + clrStripMaxY * yEdgeFracHi
      call pgrect(clrStripMinX, clrStripMaxX, yEdgeLo, yEdgeHi)
    end do

    call pgsfs(2)
    call pgsci(1)
    call pgrect(clrStripMinX, clrStripMaxX, clrStripMinY, clrStripMaxY)

    boxWidth = (boxRowMaxX - boxRowMinX) / (3.0 + 2.0&
    * (boxPadFraction) / (1.0 - boxPadFraction))
    do hueIndex = red, blu
      frac = (hueIndex - red) / 2.0
      boxMinX(hueIndex) = boxRowMinX * (1.0 - frac) + (boxRowMaxX - boxWidth)&
      * frac

      boxMaxX(hueIndex) = boxMinX(hueIndex) + boxWidth
      call drawPinningGraph(hueIndex, selectedCi, selectedHi, 1)
    end do

!    btnText(1) = 'Save cmap'
!    btnText(2) = 'Undo last'
!    btnText(3) = 'Undo all'
!    btnText(4) = 'Del point'
!    btnText(5) = '-> Default'
!    btnText(6) = 'EXIT'
btnText = '' ! vector op, default
btnText(1) = 'EXIT'
btnText(2) = '-> Default'
btnText(3) = 'Del point'

    btnHeight = (btnColMaxY - btnColMinY) / (numBtns + (numBtns - 1)&
    * (btnPadFraction) / (1.0 - btnPadFraction))
    do btnNum = 1, numBtns
      frac = (btnNum - 1) / real(numBtns - 1)
      btnMaxY(btnNum) = btnColMaxY * (1.0 - frac)  + (btnColMinY + btnHeight)&
      * frac

      btnMinY(btnNum) = btnMaxY(btnNum) - btnHeight

      call pgrect(btnColMinX, btnColMaxX, btnMinY(btnNum), btnMaxY(btnNum))
      textX = btnColMinX + 0.02
      textY = (btnMaxY(btnNum) + btnMinY(btnNum)) / 2.0
      call pgtext(textX, textY, btnText(btnNum))
    end do

  end subroutine drawInitialPlot 

  !---------------------------------------------------------------------
  subroutine drawPinningGraph(hueIndex, localSelectedCi, localSelectedHi, ciForDrawing)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: hueIndex, localSelectedCi, localSelectedHi, ciForDrawing
    integer :: clrIndex, oldCi, oldFillStyle
    real :: xVal, yVal, frac

    !-------------------------------------------------------------------
    ! Start

    call pgqci(oldCi)
    call pgqfs(oldFillStyle)
    call pgsci(ciForDrawing)
    call pgsfs(2)
    call pgrect(boxMinX(hueIndex), boxMaxX(hueIndex), boxRowMinY, boxRowMaxY)

    yVal = boxRowMinY
    frac = pins(numReservedClrs)%hue(hueIndex)
    xVal = boxMinX(hueIndex) * (1.0 - frac) + boxMaxX(hueIndex) * frac
    call pgmove(xVal, yVal)
    call pgpt1(xVal, yVal, pinSymbol)
    do clrIndex = numReservedClrs + 1, 255
    if (.not. pins(clrIndex)%isPinned(hueIndex)) cycle
      frac = (clrIndex - numReservedClrs) / real(255 - numReservedClrs)
      yVal = boxRowMinY * (1.0 - frac) + boxRowMaxY * frac
      frac = pins(clrIndex)%hue(hueIndex)
      xVal = boxMinX(hueIndex) * (1.0 - frac) + boxMaxX(hueIndex) * frac
      call pgdraw(xVal, yVal)
      call pgpt1(xVal, yVal, pinSymbol)
    end do

!    if (indexIsSelected .and. hue .eq. selectedHue) then
!      yVal = (selectedIndex - minClrIndex) / real(maxClrIndex - minClrIndex)
!      call pgpt1(rgbPinValue(hue, selectedIndex), yVal, selectionSymbol)
!    end if

    if (localSelectedCi .ne. -1 .and. localSelectedHi .ne. -1) then
      frac = (localSelectedCi - numReservedClrs) / real(255 - numReservedClrs)
      yVal = boxRowMinY * (1.0 - frac) + boxRowMaxY * frac
      frac = pins(localSelectedCi)%hue(localSelectedHi)
      xVal = boxMinX(localSelectedHi) * (1.0 - frac)&
      + boxMaxX(localSelectedHi) * frac
      call pgpt1(xVal, yVal, selectionSymbol)
    end if

    call pgsci(oldCi)
    call pgsfs(oldFillStyle)

  end subroutine drawPinningGraph 

  !---------------------------------------------------------------------
  function readButton(csrX, csrY)
  !---------------------------------------------------------------------

    implicit none

    real, intent(in) :: csrX, csrY
    integer :: btnNum, readButton

    !-------------------------------------------------------------------
    ! Start

    readButton = -1 ! default, signifies no button at all.
    if (csrX .ge. btnColMinX .and. csrX .le. btnColMaxX) then
      do btnNum = 1, numBtns
        if (csrY .ge. btnMinY(btnNum) .and. csrY .le. btnMaxY(btnNum)) then
          readButton = btnNum
      exit
        end if
      end do
    end if

  end function readButton

  !---------------------------------------------------------------------
  subroutine deletePoint(isOtherDevice, myDevId, otherDevId, status)
  !---------------------------------------------------------------------

    implicit none

    logical, intent(in) :: isOtherDevice
    integer, intent(in) :: myDevId, otherDevId
    integer, intent(out) :: status

    integer :: hueIndex

    !-------------------------------------------------------------------
    ! Start

    status = 0

    if (selectedCi .ne. -1 .and. selectedHi .ne. -1) then
      if (selectedCi .eq. numReservedClrs .or. selectedCi .eq. 255) then
        status = -3
  return
      end if

      hueIndex = selectedHi
      call drawPinningGraph(hueIndex, selectedCi, selectedHi, 0)
      ! erases the old

      pins(selectedCi)%isPinned(selectedHi) = .false.
      selectedCi = -1
      selectedHi = -1

      call drawPinningGraph(hueIndex, selectedCi, selectedHi, 1)
      ! draws in the new

      call snapColoursToPins
      call setDeviceColourMap(status)
  if (status .lt. 0) return

      if (isOtherDevice) then ! change its colour map too:
        call pgslct(otherDevId)
        call findClrLimits(status)
  if (status .lt. 0) return

        call setDeviceColourMap(status)
  if (status .lt. 0) return

        call pgslct(myDevId)
      end if
    end if

  end subroutine deletePoint

  !---------------------------------------------------------------------
  subroutine dragPin(oldCsrX, oldCsrY, newCsrX, newCsrY, isOtherDevice&
  , myDevId, otherDevId, status)
  !---------------------------------------------------------------------

    implicit none

    real, intent(inout) ::&
      oldCsrX,&
      oldCsrY
    real, intent(out) ::&
      newCsrX,&
      newCsrY
    logical, intent(in) :: isOtherDevice
    integer, intent(in) :: myDevId, otherDevId
    integer, intent(out) :: status

    integer :: hi, oldClrIndex, newClrIndex, oldSelCi, newSelCi, oldSelHi, newSelHi
    logical :: inBox, oldWasPinned
    real :: newHue
    integer ::&
      hueIndex,&
      i,&
      clrIndex,&
!      btnNum,&
      pgband
    character(len=1) ::&
      char

!------------------------------------------------------------------------
! Start:

    status = 0

    newCsrX = oldCsrX ! Defaults: so the cursor doesn't get shifted
    newCsrY = oldCsrY ! when the click is not within a box.

    inBox = .false. ! default
!    if (oldCsrY .ge. boxRowMinY .and. oldCsrY .lt. boxRowMaxY) then
    if (oldCsrY .ge. 0.0 .and. oldCsrY .lt. 1.0) then ! A tiny bit of a
    ! bodge so that the first and last pins can be moved without trouble.
      do hueIndex = red, blu
        if (oldCsrX .ge. boxMinX(hueIndex)&
        .and. oldCsrX .le. boxMaxX(hueIndex)) then
          hi = hueIndex
          inBox = .true.
      exit
        end if
      end do
    end if

    if (inBox) then
      oldClrIndex = numReservedClrs + int((256 - numReservedClrs)&
      * ((oldCsrY - boxRowMinY) / (boxRowMaxY - boxRowMinY)))
      oldClrIndex = min(255, max(numReservedClrs, oldClrIndex))

      ! Find out whether the selected point is near a previous
      ! point (=> oldWasPinned = .true.). Some leeway is applied via the
      ! parameter 'ciBoundary' so as not to make it too darn difficult
      ! to select a previous point.
      if (pins(oldClrIndex)%isPinned(hi)) then
        oldWasPinned = .true.
      else
        oldWasPinned = .false.
        do i = 1, ciBoundary
          clrIndex = oldClrIndex + i
          if (clrIndex .le. 255) then
            if (pins(clrIndex)%isPinned(hi)) then
              oldWasPinned = .true.
              oldClrIndex = clrIndex
!print *, i, clrIndex, pins(clrIndex)%isPinned(hi)
        exit
!            else
!print *, i, clrIndex, pins(clrIndex)%isPinned(hi)
            end if
          else
!print *, oldClrIndex, pins(oldClrIndex)%isPinned(hi), i, clrIndex
            status = -3
  return
!            oldWasPinned = .true.
!            oldClrIndex = 255
!        exit
          end if
          clrIndex = oldClrIndex - i
          if (clrIndex .ge. numReservedClrs) then
            if (pins(clrIndex)%isPinned(hi)) then
              oldWasPinned = .true.
              oldClrIndex = clrIndex
!print *, i, clrIndex, pins(clrIndex)%isPinned(hi)
        exit
!            else
!print *, i, clrIndex, pins(clrIndex)%isPinned(hi)
            end if
          else
!print *, oldClrIndex, pins(oldClrIndex)%isPinned(hi), i, clrIndex
            status = -3
  return
!            oldWasPinned = .true.
!            oldClrIndex = numReservedClrs
!        exit
          end if
        end do
      end if ! pins(oldClrIndex)%isPinned(hi)

      oldSelCi = selectedCi
      oldSelHi = selectedHi

      ! 'Drag' the cursor:
      status = pgband(2, 1, oldCsrX, oldCsrY, newCsrX, newCsrY, char)

!      btnNum = readButton(newCsrX, newCsrY)
!      if (btnNum .eq. 2) then ! deselect point, then do nothing further
!      ! with the drag, just exit.
!        inBox = .false.
!        newSelCi = oldSelCi
!
!      else ! btnNum .ne. 2, work out the new hue and colour index:
        if (oldClrIndex .eq. 255 .or. oldClrIndex .eq. numReservedClrs) then
          newClrIndex = oldClrIndex
!print *, 'old = new:', oldClrIndex, newClrIndex
!          newSelCi = oldSelCi ! should be -1, these end points can't be
!          ! selected (this is to save them from being deleted).

        else
          newClrIndex = numReservedClrs + int((256 - numReservedClrs)&
          * ((newCsrY - boxRowMinY) / (boxRowMaxY - boxRowMinY)))
          newClrIndex = min(255, max(numReservedClrs, newClrIndex))
!          newSelCi = newClrIndex
!print *, 'old, new:', oldClrIndex, newClrIndex
        end if
!        selectedCi = newSelCi

        ! But the pinned clr index (whether new or old) shouldn't be able
        ! to move to a clr index on the other side of the next pinned
        ! point; it has to keep to its own space. So find the limits in
        ! the direction of movement:
        if (newClrIndex .gt. oldClrIndex) then
          do clrIndex = oldClrIndex + 1, newClrIndex
            if (pins(clrIndex)%isPinned(hi)) then
              newClrIndex = clrIndex - 1
          exit
            end if
          end do
        else if (newClrIndex .lt. oldClrIndex) then
          do clrIndex = oldClrIndex - 1, newClrIndex, -1
            if (pins(clrIndex)%isPinned(hi)) then
              newClrIndex = clrIndex + 1
          exit
            end if
          end do
        end if ! we don't worry about equality.

        ! Now work out the new hue:
        newHue = (newCsrX - boxMinX(hi)) / (boxMaxX(hi) - boxMinX(hi))
        newHue = min(1.0, max(0.0, newHue))

        if (oldClrIndex .eq. 255 .or. oldClrIndex .eq. numReservedClrs) then
          newSelCi = oldSelCi ! should be -1: these end points can't be
          newSelHi = oldSelHi ! selected (this is to save them from being
          ! deleted).
          ! 
        else
          newSelCi = newClrIndex
          newSelHi = hi
        end if
        selectedCi = newSelCi
        selectedHi = newSelHi

!      end if ! btnNum .eq. 2
!      selectedCi = newSelCi

      call drawPinningGraph(hi, oldSelCi, oldSelHi, 0)
      ! erases the old

      if (oldWasPinned) then
        pins(oldClrIndex)%isPinned(hi) = .false. ! effectively,
        ! this moves the old pin to the new position rather than
        ! introducing a new pin.
      end if
      pins(newClrIndex)%isPinned(hi) = .true.
      pins(newClrIndex)%hue(hi) = newHue
      call drawPinningGraph(hi, newSelCi, newSelHi, 1)
      ! draws in the new

      call snapColoursToPins
      call setDeviceColourMap(status)
  if (status .lt. 0) return

      if (isOtherDevice) then ! change its colour map too:
        call pgslct(otherDevId)
        call findClrLimits(status)
  if (status .lt. 0) return

        call setDeviceColourMap(status)
  if (status .lt. 0) return

        call pgslct(myDevId)
      end if
    end if ! inBox

  end subroutine dragPin

  !---------------------------------------------------------------------
  subroutine plotGTIs(unsortedGtiArray, returnedAndedTimes&
  , returnedOredTimes, device, label, doNeatTimeLabel, status)
  !---------------------------------------------------------------------

    implicit none

    type(gtiType), intent(in)                        :: unsortedGtiArray(:)
    real(kind(0d0)), dimension(:), pointer, optional :: returnedAndedTimes,&
                                                        returnedOredTimes
    character(*),  intent(in),              optional :: device, label
    logical,       intent(in),              optional :: doNeatTimeLabel
    integer,       intent(out),             optional :: status

    type(gtiType) :: gtiArray(size(unsortedGtiArray))
    real(kind(0d0)) :: andedTimes(size(unsortedGtiArray)), oredTimes(size(unsortedGtiArray)), startTime, previousTime, timeOffset
    real, dimension(:), allocatable :: duration
    integer :: numTimes, numAndedTimes, numOredTimes, numFiles,&
      deviceId, pgopen, fileNum, i, j, localStatus,&
      numSimultaneousGtis, indices(size(unsortedGtiArray))
    real :: minT, maxT, minY, maxY, tEdgeWidth, yEdgeWidth,&
      stripHeight, loY, hiY, startT, endT, frac, andDuration, orDuration
    real :: r = 0.8 ! ratio of strip height to sum of height + gap between strips.
    logical :: withinGti
    character(255) :: localDevice, localLabel, timeLabel
    character(4) :: labelStr
    character(6) :: fracStr

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    numTimes = size(unsortedGtiArray)
    call mySort(unsortedGtiArray%time, indices, status)
    do i = 1, numTimes
      gtiArray(i) = unsortedGtiArray(indices(i))
    end do

    numFiles = maxval(gtiArray%fileNum)
    allocate(duration(numFiles))
    duration = 0.0
    andDuration = 0.0
    orDuration = 0.0

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

    if (present(label)) then
      localLabel = label
    else
      localLabel = ''
    end if

! Do checks:

!do i = 1, numTimes
!  print *, i, gtiArray(i)%time, gtiArray(i)%fileNum, gtiArray(i)%isStart
!end do

    ! Check order, and that stops alternate with starts:
    do fileNum = 1, numFiles
      previousTime = minval(gtiArray%time)
      withinGti = .false.
      do i = 1, numTimes
      if (gtiArray(i)%fileNum .ne. fileNum) cycle
!print *, fileNum, previousTime, withinGti, i, gtiArray(i)%time
        if (gtiArray(i)%time .lt. previousTime) then
          if (present(status)) status = -1
  return
        end if
        previousTime = gtiArray(i)%time

        if (withinGti) then
          if (gtiArray(i)%isStart) then
            if (present(status)) status = -2
  return
          end if

        else
          if (.not. gtiArray(i)%isStart) then
            if (present(status)) status = -3
  return
          end if
        end if
        withinGti = .not. withinGti
      end do
    end do

    if (.not. gtiArray(1)%isStart .or. gtiArray(numTimes)%isStart) then
      if (present(status)) status = -100 ! actually a bug - shouldn't happen.
  return
    end if

! Calculate time offset if necessary:

    if (doNeatTimeLabel) then
      timeOffset = floor(minval(gtiArray%time) / 1000.0) * 1000.0
      gtiArray%time = gtiArray%time - timeOffset
!      write(timeLabel, '(a, es11.5, a)') 'Time - ', timeOffset, ' (sec)'
      write(timeLabel, '(a, i10, a)') 'Time - ', int(timeOffset), ' (sec)'
    else
      timeLabel = 'Time (sec)'
    end if

! Now load the separate arrays:

    numAndedTimes = 0
    numOredTimes  = 0
    numSimultaneousGtis = 0
    do i = 1, numTimes
      if (gtiArray(i)%isStart) then
        numSimultaneousGtis = numSimultaneousGtis + 1
        if (numSimultaneousGtis .gt. numFiles) then
          if (present(status)) status = -200 ! actually a bug - shouldn't
          ! happen.
  return
        end if

        if (numSimultaneousGtis .eq. 1) then
          numOredTimes = numOredTimes + 1
          oredTimes(numOredTimes) = gtiArray(i)%time

        else if (numSimultaneousGtis .eq. numFiles) then
          numAndedTimes = numAndedTimes + 1
          andedTimes(numAndedTimes) = gtiArray(i)%time
        end if

      else
        if (numSimultaneousGtis .eq. 1) then
          numOredTimes = numOredTimes + 1
          oredTimes(numOredTimes) = gtiArray(i)%time

        else if (numSimultaneousGtis .eq. numFiles) then
          numAndedTimes = numAndedTimes + 1
          andedTimes(numAndedTimes) = gtiArray(i)%time
        end if

        numSimultaneousGtis = numSimultaneousGtis - 1
        if (numSimultaneousGtis .lt. 0) then
          if (present(status)) status = -300 ! actually a bug - shouldn't
          ! happen.
  return
        end if
      end if
    end do

    i = 1
    do
      if (andedTimes(i) .eq. andedTimes(i + 1)) then
      ! Delete the pair of equal times:
        do j = i, numAndedTimes - 2
          andedTimes(j) = andedTimes(j + 2)
        end do
        numAndedTimes = numAndedTimes - 2
      else
        i = i + 1
      end if
    if (i .gt. numAndedTimes - 1) exit
    end do

    i = 1
    do
      if (oredTimes(i) .eq. oredTimes(i + 1)) then
      ! Delete the pair of equal times:
        do j = i, numOredTimes - 2
          oredTimes(j) = oredTimes(j + 2)
        end do
        numOredTimes = numOredTimes - 2
      else
        i = i + 1
      end if
    if (i .gt. numOredTimes - 1) exit
    end do

    if (present(returnedAndedTimes)) then
      allocate(returnedAndedTimes(numAndedTimes))
      returnedAndedTimes = andedTimes(1:numAndedTimes)
    end if
    if (present(returnedOredTimes)) then
      allocate(returnedOredTimes(numOredTimes))
      returnedOredTimes  = oredTimes( 1:numOredTimes)
    end if

! Calculate durations:

    do fileNum = 1, numFiles
      withinGti = .false.
      do i = 1, numTimes
      if (gtiArray(i)%fileNum .ne. fileNum) cycle
        if (withinGti) then
          duration(fileNum) = duration(fileNum) + real(gtiArray(i)%time&
          - startTime)
        else
          startTime = gtiArray(i)%time
        end if
        withinGti = .not. withinGti
      end do
    end do

    withinGti = .false.
    do i = 1, numAndedTimes
      if (withinGti) then
        andDuration = andDuration + real(andedTimes(i) - startTime)
      else
        startTime = andedTimes(i)
      end if
      withinGti = .not. withinGti
    end do

    withinGti = .false.
    do i = 1, numOredTimes
      if (withinGti) then
        orDuration = orDuration + real(oredTimes(i) - startTime)
      else
        startTime = oredTimes(i)
      end if
      withinGti = .not. withinGti
    end do

! Now plot:

    minT = minval(gtiArray%time)
    maxT = maxval(gtiArray%time)
    minY = 0.0
    maxY = real(numFiles + 3)
    stripHeight = (maxY - minY) / (numFiles + 3 + (numFiles + 2) * (1.0 - r)&
    / r) 

    deviceId = pgopen(trim(localDevice))
    call pgsvp(0.07, 0.999, 0.1, 0.90)
!***** Use standard values??

    tEdgeWidth = 0.1 * (maxT - minT)
    yEdgeWidth = 0.05 * (maxY - minY)
    call pgswin(minT - 0.1 * (maxT - minT), maxT + 0.13 * (maxT - minT)&
    , minY - yEdgeWidth, maxY + yEdgeWidth)

    call pgsfs(1)

    do fileNum = 1, numFiles
      frac = (fileNum - 1) / real(numFiles + 3 - 1)
      loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
      hiY = loY + stripHeight

!*** change to cater for devices with few colours!
      call pgscr(4, 0.50, 0.32, 0.25)
      withinGti = .false.
      startT = minT
      do i = 1, numTimes
      if (gtiArray(i)%fileNum .ne. fileNum) cycle
        endT = gtiArray(i)%time
        if (withinGti) then
          if (gtiArray(i)%isStart) then
            if (present(status)) status = -4
  return
          end if
          call pgsci(4)

        else
          if (.not. gtiArray(i)%isStart) then
            if (present(status)) status = -5
  return
          end if
          call pgsci(0)
        end if
        call pgrect(startT, endT, loY, hiY)
        startT = endT
        withinGti = .not. withinGti
      end do

      call pgsci(1)
      write(labelStr, '(i4)') fileNum
      call pgtext(minT - tEdgeWidth, loY, labelStr)
      write(fracStr, '(a1, f5.3)') ' ', duration(fileNum) /&
      (oredTimes(numOredTimes) - oredTimes(1))
      call pgtext(maxT, loY, fracStr)
    end do

!    call pgsci(2)
    frac = (numFiles + 1) / real(numFiles + 3 - 1)
    loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
    hiY = loY + stripHeight
    withinGti = .false.
    startT = minT
    do i = 1, numAndedTimes
      endT = andedTimes(i)
      if (withinGti) then
        call pgsci(2)

      else
        call pgsci(0)
      end if
      call pgrect(startT, endT, loY, hiY)
      startT = endT
      withinGti = .not. withinGti
    end do
    call pgsci(1)
    call pgtext(minT - tEdgeWidth, loY, ' AND')
    write(fracStr, '(a1, f5.3)') ' ', andDuration / (oredTimes(numOredTimes)&
    - oredTimes(1))
    call pgtext(maxT, loY, fracStr)

!    call pgsci(3)
    call pgscr(3, 0.0, 0.7, 0.0)
    frac = (numFiles + 2) / real(numFiles + 3 - 1)
    loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
    hiY = loY + stripHeight
    withinGti = .false.
    startT = minT
    do i = 1, numOredTimes
      endT = oredTimes(i)
      if (withinGti) then
        call pgsci(3)

      else
        call pgsci(0)
      end if
      call pgrect(startT, endT, loY, hiY)
      startT = endT
      withinGti = .not. withinGti
    end do
    call pgsci(1)
    call pgtext(minT - tEdgeWidth, loY, ' OR')
    write(fracStr, '(a1, f5.3)') ' ', orDuration / (oredTimes(numOredTimes)&
    - oredTimes(1))
    call pgtext(maxT, loY, fracStr)

    call pgsci(1)
    call pgsfs(2)
    call pgbox('BCNT', 0., 0, 'BC', 0., 0)
    call pglab(trim(timeLabel), '', trim(localLabel))
    call pgclos

    if (present(status)) status = localStatus

  end subroutine plotGTIs

  !---------------------------------------------------------------------
  subroutine mySort(vector, indices, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(0d0)), intent(in) :: vector(:)
    integer, intent(out) :: indices(:), status
integer :: i

    real(kind(0d0)) :: tempVector(size(vector))

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    if (size(indices) .ne. size(vector)) then
      status = -1
  return
    end if

    tempVector = vector
    call quick_sort(tempVector, indices)

!do i = 1, size(vector)
!print *, i, indices(i), vector(i), vector(indices(i)), tempVector(i), !tempVector(indices(i))
!end do
!call error('','')

  end subroutine mySort

  !---------------------------------------------------------------------
  subroutine plotGTIs_med(times, fileNums, isStart, returnedAndedTimes&
  , returnedOredTimes, device, status)
  !---------------------------------------------------------------------

    implicit none

    real(kind(0d0)), intent(in) :: times(:)
    integer, intent(in) :: fileNums(:)
    logical, intent(in) :: isStart(:)
    real(kind(0d0)), dimension(:), pointer :: returnedAndedTimes, returnedOredTimes
    character(*), intent(in), optional :: device
    integer, intent(out), optional :: status

    real(kind(0d0)) :: andedTimes(size(times)), oredTimes(size(times)), startTime
    real, dimension(:), allocatable :: duration
    integer :: numTimes, numAndedTimes, numOredTimes, numFiles,&
      deviceId, pgopen, fileNum, i, j, localStatus,&
      numSimultaneousGtis
    real :: minT, maxT, minY, maxY, tEdgeWidth, yEdgeWidth,&
      stripHeight, loY, hiY, loT, hiT, frac, andDuration
    real :: r = 0.8 ! ratio of strip height to sum of height + gap between strips.
    logical :: withinGti
    character(255) :: localDevice
    character(4) :: labelStr
    character(6) :: fracStr

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    numTimes = size(times)
    numFiles = maxval(fileNums)
    allocate(duration(numFiles))
    duration = 0.0
    andDuration = 0.0

    if (present(device)) then
      localDevice = trim(device)
    else
      localDevice = defaultDevice
    end if

! Do checks:

    ! Check order:
    do i = 1, numTimes - 1
if (times(i) .gt. times(i + 1)) call error('outOfOrder','')
    end do

    ! Check stops follow starts:
    do fileNum = 1, numFiles
      withinGti = .false.
      do i = 1, numTimes
      if (fileNums(i) .ne. fileNum) cycle
        if (withinGti) then
          if (isStart(i)) then
print *, 'out-of-place start, file num', fileNum, 'entry', i
call error('','')
          end if
        else
          if (.not. isStart(i)) then
print *, 'out-of-place  stop, file num', fileNum, 'entry', i
call error('','')
          end if
        end if
        withinGti = .not. withinGti
      end do
    end do

    if (.not. isStart(1) .or. isStart(numTimes)) then
call error('bug0', '')
    end if

call message(AppMsg, Verbose, 'Constructing AND and OR of GTIs')

    numAndedTimes = 0
    numOredTimes  = 0
    numSimultaneousGtis = 0
    do i = 1, numTimes
      if (isStart(i)) then
        numSimultaneousGtis = numSimultaneousGtis + 1
if (numSimultaneousGtis .gt. numFiles) call error('bug1','')
!write(*, '(i5, i4, l3, i4, d25.17)') i, fileNums(i)&
!, isStart(i), numSimultaneousGtis, times(i)
        if (numSimultaneousGtis .eq. 1) then
          numOredTimes = numOredTimes + 1
          oredTimes(numOredTimes) = times(i)

        else if (numSimultaneousGtis .eq. numFiles) then
          numAndedTimes = numAndedTimes + 1
          andedTimes(numAndedTimes) = times(i)
        end if

      else
        if (numSimultaneousGtis .eq. 1) then
          numOredTimes = numOredTimes + 1
          oredTimes(numOredTimes) = times(i)

        else if (numSimultaneousGtis .eq. numFiles) then
          numAndedTimes = numAndedTimes + 1
          andedTimes(numAndedTimes) = times(i)
        end if

        numSimultaneousGtis = numSimultaneousGtis - 1
if (numSimultaneousGtis .lt. 0) call error('bug2','')
!write(*, '(i5, i4, l3, i4, d25.17)') i, fileNums(i)&
!, isStart(i), numSimultaneousGtis, times(i)
      end if
    end do

    i = 1
    do
      if (andedTimes(i) .eq. andedTimes(i + 1)) then
      ! Delete the pair of equal times:
        do j = i, numAndedTimes - 2
          andedTimes(j) = andedTimes(j + 2)
        end do
        numAndedTimes = numAndedTimes - 2
      else
        i = i + 1
      end if
    if (i .gt. numAndedTimes - 1) exit
    end do

    i = 1
    do
      if (oredTimes(i) .eq. oredTimes(i + 1)) then
      ! Delete the pair of equal times:
        do j = i, numOredTimes - 2
          oredTimes(j) = oredTimes(j + 2)
        end do
        numOredTimes = numOredTimes - 2
      else
        i = i + 1
      end if
    if (i .gt. numOredTimes - 1) exit
    end do

allocate(returnedAndedTimes(numAndedTimes), returnedOredTimes(numOredTimes))
returnedAndedTimes = andedTimes(1:numAndedTimes)
returnedOredTimes  = oredTimes( 1:numOredTimes)

! Calculate durations:

    do fileNum = 1, numFiles
      withinGti = .false.
      do i = 1, numTimes
      if (fileNums(i) .ne. fileNum) cycle
        if (withinGti) then
          duration(fileNum) = duration(fileNum) + real(times(i) - startTime)
        else
          startTime = times(i)
        end if
        withinGti = .not. withinGti
      end do
    end do

    withinGti = .false.
    do i = 1, numAndedTimes
      if (withinGti) then
        andDuration = andDuration + real(andedTimes(i) - startTime)
      else
        startTime = andedTimes(i)
      end if
      withinGti = .not. withinGti
    end do

! Now plot:

    minT = minval(times)
    maxT = maxval(times)
    minY = 0.0
    maxY = real(numFiles + 3)
    stripHeight = (maxY - minY) / (numFiles + 3 + (numFiles + 2) * (1.0 - r) / r) 

    deviceId = pgopen(trim(localDevice))
    call pgsvp(0.07, 0.999, 0.1, 0.90)

    tEdgeWidth = 0.1 * (maxT - minT)
    yEdgeWidth = 0.05 * (maxY - minY)
    call pgswin(minT - tEdgeWidth, maxT + tEdgeWidth&
    , minY - yEdgeWidth, maxY + yEdgeWidth)

!    call pgsci(2)
    call pgsfs(1)

    do fileNum = 1, numFiles
      frac = (fileNum - 1) / real(numFiles + 3 - 1)
      loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
      hiY = loY + stripHeight

      call pgsci(4)
!      call pgscr(4, 0.6, 0.35, 0.2)
      call pgscr(4, 0.50, 0.32, 0.25)
      withinGti = .false.
      do i = 1, numTimes
      if (fileNums(i) .ne. fileNum) cycle
        if (withinGti) then
if (isStart(i)) call error('badStart','')
          hiT = times(i)
          call pgrect(loT, hiT, loY, hiY)

        else
if (.not. isStart(i)) call error('badStop','')
          loT = times(i)
        end if
        withinGti = .not. withinGti
      end do

      call pgsci(1)
      write(labelStr, '(i4)') fileNum
      call pgtext(minT - tEdgeWidth, loY, labelStr)
!print *, ' ', duration(fileNum) /&
!      (oredTimes(numOredTimes) - oredTimes(1))
      write(fracStr, '(a1, f5.3)') ' ', duration(fileNum) /&
      (oredTimes(numOredTimes) - oredTimes(1))
!print *, fracStr
      call pgtext(maxT, loY, fracStr)
    end do

    call pgsci(2)
    frac = (numFiles + 1) / real(numFiles + 3 - 1)
    loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
    hiY = loY + stripHeight
    withinGti = .false.
    do i = 1, numAndedTimes
      if (withinGti) then
        hiT = andedTimes(i)
        call pgrect(loT, hiT, loY, hiY)

      else
        loT = andedTimes(i)
      end if
      withinGti = .not. withinGti
    end do
    call pgsci(1)
    call pgtext(minT - tEdgeWidth, loY, ' AND')
    write(fracStr, '(a1, f5.3)') ' ', andDuration / (oredTimes(numOredTimes)&
    - oredTimes(1))
    call pgtext(maxT, loY, fracStr)

    call pgsci(3)
    call pgscr(3, 0.0, 0.7, 0.0)
    frac = (numFiles + 2) / real(numFiles + 3 - 1)
    loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
    hiY = loY + stripHeight
    withinGti = .false.
    do i = 1, numOredTimes
      if (withinGti) then
        hiT = oredTimes(i)
        call pgrect(loT, hiT, loY, hiY)

      else
        loT = oredTimes(i)
      end if
      withinGti = .not. withinGti
    end do
    call pgsci(1)
    call pgtext(minT - tEdgeWidth, loY, ' OR')

    call pgsci(1)
    call pgsfs(2)
    call pgbox('BCNT', 0., 0, 'BC', 0., 0)
    call pgclos

    if (present(status)) status = localStatus

  end subroutine plotGTIs_med

  !---------------------------------------------------------------------
  subroutine plotGTIs_old(times, fileNums, isStart, andedTimes, oredTimes, device)
  !---------------------------------------------------------------------

    implicit none

!    type(gtiType), intent(in) :: times(:), andedTimes(:), oredTimes(:)
    real(kind(0d0)), intent(in) :: times(:), andedTimes(:), oredTimes(:)
    character(*), intent(in) :: device
    integer, intent(in) :: fileNums(:)
    logical, intent(in) :: isStart(:)
    integer :: numTimes, numAndedTimes, numOredTimes, numFiles, deviceId, pgopen, fileNum, i
    real :: minT, maxT, minY, maxY, tEdgeWidth, yEdgeWidth, stripHeight, loY, hiY, loT, hiT, frac
    real :: r = 0.8 ! ratio of strip height to sum of height + gap between strips
    logical :: withinGti

    !-------------------------------------------------------------------
    ! Start

    numTimes = size(times)
    numAndedTimes = size(andedTimes)
    numOredTimes = size(oredTimes)

    numFiles = maxval(fileNums)

    minT = minval(times)
    maxT = maxval(times)
    minY = 0.0
    maxY = real(numFiles + 3)
    stripHeight = (maxY - minY) / (numFiles + 3 + (numFiles + 2) * (1.0 - r) / r) 

    deviceId = pgopen(trim(device))
    call pgsvp(0.07, 0.999, 0.1, 0.90)

    tEdgeWidth = 0.05 * (maxT - minT)
    yEdgeWidth = 0.05 * (maxY - minY)
    call pgswin(minT - tEdgeWidth, maxT + tEdgeWidth&
    , minY - yEdgeWidth, maxY + yEdgeWidth)

    call pgsci(2)
    call pgsfs(1)

    do fileNum = 1, numFiles
      frac = (fileNum - 1) / real(numFiles + 3 - 1)
      loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
      hiY = loY + stripHeight

      withinGti = .false.
      do i = 1, numTimes
      if (fileNums(i) .ne. fileNum) cycle
        if (withinGti) then
if (isStart(i)) call error('badStart','')
          hiT = times(i)
          call pgrect(loT, hiT, loY, hiY)

        else
if (.not. isStart(i)) call error('badStop','')
          loT = times(i)
        end if
        withinGti = .not. withinGti
      end do
    end do

    call pgsci(3)
    frac = (numFiles + 1) / real(numFiles + 3 - 1)
    loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
    hiY = loY + stripHeight
    withinGti = .false.
    do i = 1, numAndedTimes
      if (withinGti) then
        hiT = andedTimes(i)
        call pgrect(loT, hiT, loY, hiY)

      else
        loT = andedTimes(i)
      end if
      withinGti = .not. withinGti
    end do

    call pgsci(4)
    frac = (numFiles + 2) / real(numFiles + 3 - 1)
    loY = (maxY - stripHeight) * (1.0 - frac) + minY * frac
    hiY = loY + stripHeight
    withinGti = .false.
    do i = 1, numOredTimes
      if (withinGti) then
        hiT = oredTimes(i)
        call pgrect(loT, hiT, loY, hiY)

      else
        loT = oredTimes(i)
      end if
      withinGti = .not. withinGti
    end do

    call pgsci(1)
    call pgsfs(2)
    call pgbox('BCNT', 0., 0, 'BC', 0., 0)
    call pgclos

  end subroutine plotGTIs_old

  !---------------------------------------------------------------------
  subroutine mouse(deviceId)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: deviceId
    integer :: deviceStatus, csrStatus, pgcurs
    real :: x, y
    character(1) :: char

    !-------------------------------------------------------------------
    ! Start

    deviceStatus = queryDeviceStatus(deviceId)
    if (deviceStatus .le. 1) then
  return
    else if (deviceStatus .eq. 2) then
      call pgslct(deviceId)
    end if

    csrStatus = 1
    do while(csrStatus .eq. 1)
      csrStatus = pgcurs(x, y, char)
      if (char .ne. 'A' .and. char .ne. 'D' .and. char .ne. 'X') then
  return
      end if
print *, char
    end do

  end subroutine mouse

  !---------------------------------------------------------------------
  subroutine rotatingDisplayVector(points, numStepsPerQuadrant, obsThetaDeg&
  , obsPhiDeg, ranges, device, withGrid, status)
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    real,         intent(in)            :: points(:, :)
    integer,      intent(in),  optional :: numStepsPerQuadrant
    real,         intent(in),  optional :: obsThetaDeg,&
                                           obsPhiDeg,&
                                           ranges(3, 2)
    character(*), intent(in),  optional :: device
    logical,      intent(in),  optional :: withGrid
    integer,      intent(out), optional :: status

    real ::&
      localObsThetaDeg,&
      localRanges(3, 2),&
      localPoints(size(points, 1), 1, size(points, 2)),&
      localObsPhiDeg
    character(255) ::&
      localDevice
    integer ::&
      localStatus,&
      coord,&
      localNumStepsPerQuadrant
    logical ::&
      localWithGrid

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    localPoints(:, 1, :) = points

    call initDefaultObs

    if (present(numStepsPerQuadrant)) then
      localNumStepsPerQuadrant = numStepsPerQuadrant
    else
      localNumStepsPerQuadrant = 45
    end if

    if (present(obsThetaDeg)) then
      localObsThetaDeg = obsThetaDeg
      do while (localObsThetaDeg .lt. 0.0)
        localObsThetaDeg = localObsThetaDeg + 360.0
      end do
      do while (localObsThetaDeg .ge. 360.0)
        localObsThetaDeg = localObsThetaDeg - 360.0
      end do
    else
      localObsThetaDeg = defaultObs%locTheta / myPiOn180
    end if

    if (present(obsPhiDeg)) then
      localObsPhiDeg = obsPhiDeg
    else
      localObsPhiDeg = defaultObs%locPhi / myPiOn180
    end if

    if (present(ranges)) then
      localRanges = ranges
    else
      do coord = 1, 3
        localRanges(coord, 1) = minval(points(:, coord))
        localRanges(coord, 2) = maxval(points(:, coord))
      end do
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(withGrid)) then
      localWithGrid = withGrid
    else
      localWithGrid = .false.
    end if

    call rotatingDisplayKernel(localPoints, localNumStepsPerQuadrant&
    , localObsThetaDeg, localObsPhiDeg, localRanges, localDevice&
    , localWithGrid, localStatus)

    if(present(status)) status = localStatus

  end subroutine rotatingDisplayVector

  !---------------------------------------------------------------------
  subroutine rotatingDisplayArray(points, numStepsPerQuadrant, obsThetaDeg&
  , obsPhiDeg, ranges, device, withGrid, status)
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    real,         intent(in)            :: points(:, :, :)
    integer,      intent(in),  optional :: numStepsPerQuadrant
    real,         intent(in),  optional :: obsThetaDeg,&
                                           obsPhiDeg,&
                                           ranges(3, 2)
    character(*), intent(in),  optional :: device
    logical,      intent(in),  optional :: withGrid
    integer,      intent(out), optional :: status

    real ::&
      localObsThetaDeg,&
      localRanges(3, 2),&
      localPoints(size(points, 1), size(points, 2), size(points, 3)),&
      localObsPhiDeg
    character(255) ::&
      localDevice
    integer :: localStatus,&
      coord,&
      localNumStepsPerQuadrant
    logical ::&
      localWithGrid

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    localPoints = points

    call initDefaultObs

    if (present(numStepsPerQuadrant)) then
      localNumStepsPerQuadrant = numStepsPerQuadrant
    else
      localNumStepsPerQuadrant = 45
    end if

    if (present(obsThetaDeg)) then
      localObsThetaDeg = obsThetaDeg
      do while (localObsThetaDeg .lt. 0.0)
        localObsThetaDeg = localObsThetaDeg + 360.0
      end do
      do while (localObsThetaDeg .ge. 360.0)
        localObsThetaDeg = localObsThetaDeg - 360.0
      end do
    else
      localObsThetaDeg = defaultObs%locTheta / myPiOn180
    end if

    if (present(obsPhiDeg)) then
      localObsPhiDeg = obsPhiDeg
    else
      localObsPhiDeg = defaultObs%locPhi / myPiOn180
    end if

    if (present(ranges)) then
      localRanges = ranges
    else
      do coord = 1, 3
        localRanges(coord, 1) = minval(points(:, :, coord))
        localRanges(coord, 2) = maxval(points(:, :, coord))
      end do
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(withGrid)) then
      localWithGrid = withGrid
    else
      localWithGrid = .false.
    end if

    call rotatingDisplayKernel(localPoints, localNumStepsPerQuadrant&
    , localObsThetaDeg, localObsPhiDeg, localRanges, localDevice&
    , localWithGrid, localStatus)

    if(present(status)) status = localStatus

  end subroutine rotatingDisplayArray

  !---------------------------------------------------------------------
  subroutine rotatingDisplayKernel(points, numStepsPerQuadrant, obsThetaDeg&
  , obsPhiDeg, ranges, device, withGrid, status)
  !---------------------------------------------------------------------
!*** sort out status numbers.

    implicit none

    !-------------------------------------------------------------------
    ! Start

    real,         intent(in)  :: points(:, :, :)
    integer,      intent(in)  :: numStepsPerQuadrant
    real,         intent(in)  :: obsThetaDeg,&
                                 obsPhiDeg,&
                                 ranges(3, 2)
    character(*), intent(in)  :: device
    logical,      intent(in)  :: withGrid
    integer,      intent(out) :: status

    real ::&
      boxMin(3), boxMax(3), radius2, maxRadius2,&
      projMaxDiameter, angle, restViewFrame(3, 3), restObsLocation(3)
    real, allocatable :: cosAngle(:), sinAngle(:)
    integer :: pgopen, xsize, ysize, coord, foreignStatus, deviceId, xi, yi,&
      zi, i, numSteps, step, edgeCI(3, 2, 2), pointsCI
    type(obsType) :: myObs
! has real :: location(3), locRadius, locTheta, locPhi, viewFrame(3, 3), viewTheta, viewPhi, viewTwist, projOffsets(2), projScale
    type(pointType) ::&
      box(2, 2, 2),& ! the bounding box of the displayed surface.
      normalisedPoints(size(points, 1), size(points, 2))
! has real :: solid(3), proj(2)
    logical :: edgeIsBkg(3, 2, 2)

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    xsize = size(points, 1)
    ysize = size(points, 2)
    if (size(points, 3) .ne. 3) then
      status = -1
  return
    end if

    edgeCI = 1 ! array op, default.
    edgeCI(1, 1, 1) = 2
    edgeCI(2, 1, 1) = 3
    edgeCI(3, 1, 1) = 4
    pointsCI = 5

    myObs%locRadius = defaultObs%locRadius

    numSteps = 4 * numStepsPerQuadrant

    myObs%locTheta = obsThetaDeg * myPiOn180
    myObs%locPhi = obsPhiDeg * myPiOn180
    if (myObs%locPhi .lt. -myPiOn2 .or. myObs%locPhi .gt. myPiOn2) then
      status = -1
  return
    end if

    boxMin = -0.5
    boxMax = 0.5

    box(1, 1:2, 1:2)%solid(1) = boxMin(1)
    box(2, 1:2, 1:2)%solid(1) = boxMax(1)
    box(1:2, 1, 1:2)%solid(2) = boxMin(2)
    box(1:2, 2, 1:2)%solid(2) = boxMax(2)
    box(1:2, 1:2, 1)%solid(3) = boxMin(3)
    box(1:2, 1:2, 2)%solid(3) = boxMax(3)

    call fillOutObs(myObs)
    myObs%projOffsets = 0.0
    ! Find max radius of box corners:
    maxRadius2&
    = box(1, 1, 1)%solid(1) * box(1, 1, 1)%solid(1)&
    + box(1, 1, 1)%solid(2) * box(1, 1, 1)%solid(2)&
    + box(1, 1, 1)%solid(3) * box(1, 1, 1)%solid(3)
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          radius2&
          = box(xi, yi, zi)%solid(1) * box(xi, yi, zi)%solid(1)&
          + box(xi, yi, zi)%solid(2) * box(xi, yi, zi)%solid(2)&
          + box(xi, yi, zi)%solid(3) * box(xi, yi, zi)%solid(3)
          maxRadius2 = max(maxRadius2, radius2)
        end do
      end do
    end do
    if (myObs%locRadius .le. maxRadius2) then
      status = -2
  return
    end if

    projMaxDiameter = 2.0 * sqrt(maxRadius2&
    / (myObs%locRadius * myObs%locRadius - maxRadius2))
!    myObs%projScale = 0.9 / projMaxDiameter
    myObs%projScale = 1.0 / projMaxDiameter

    restViewFrame = myObs%viewFrame
    restObsLocation = myObs%location

    do coord = 1, 3
      if (ranges(coord, 2) - ranges(coord, 1) .le. 0.0) then
!****** set warning status
        normalisedPoints%solid(coord) = 0.5
      else
        normalisedPoints%solid(coord)&
        = (points(:, :, coord) - ranges(coord, 1))&
        / (ranges(coord, 2) - ranges(coord, 1))
      end if

      normalisedPoints%solid(coord)&
      = boxMin(coord) * (1.0 - normalisedPoints%solid(coord))&
      + boxMax(coord) * normalisedPoints%solid(coord)
    end do

    deviceId = pgopen(trim(device))
    call findClrLimits(foreignStatus)
!***** integrate foreignStatus and status
  if (status .lt. 0) return

    call pgsvp(0.0, 0.999, 0.0, 0.999)
    call pgwnad(-0.5, 0.5, -0.5, 0.5)
!    if (.not. coloursAlreadyLoaded) call loadDefaultColourMap
!    call setDeviceColourMap(foreignStatus)
!***** integrate foreignStatus and status
  if (status .lt. 0) return

    call pgscr(pointsCI, 0.5, 0.5, 0.5)

    allocate(cosAngle(numSteps), sinAngle(numSteps))
    do i = 1, numSteps
      angle = myTwoPi * ((i - 1) / real(numSteps))
      cosAngle(i) = cos(angle)
      sinAngle(i) = sin(angle)
    end do

    call pgsfs(2) ! unfilled

!.......................................................................
! Main routine:

    step = 1
    do! while (.not. stopSignalFound())

      ! Increment rotation of observer vector:
      step = step + 1
      if (step .gt. numSteps) step = 1
      myObs%viewFrame(1, :) = restViewFrame(1, :) * cosAngle(step)&
      - restViewFrame(3, :) * sinAngle(step)
      myObs%viewFrame(3, :) = restViewFrame(3, :) * cosAngle(step)&
      + restViewFrame(1, :) * sinAngle(step)
      ! Rotation axis is the y-axis of the rest view frame, so this stays
      ! unchanged.

      myObs%location = myObs%viewFrame(3, :) * myObs%locRadius

      ! Calculate projected point positions:
      do xi = 1, 2
        do yi = 1, 2
          do zi = 1, 2
            call project_new(box(xi, yi, zi)%solid, myObs&
            , box(xi, yi, zi)%proj, foreignStatus)
!***** integrate foreignStatus and localStatus
  if (status .lt. 0) return
          end do
        end do
      end do
      do xi = 1, xsize
        do yi = 1, ysize
          call project_new(normalisedPoints(xi, yi)%solid, myObs&
          , normalisedPoints(xi, yi)%proj, foreignStatus)
!***** integrate foreignStatus and status
  if (status .lt. 0) return
        end do
      end do

      ! Decide which are bkg and which frg box edges:
      call decideBkgEdges(myObs%location, boxMin, boxMax, edgeIsBkg)
!edgeIsBkg = .false.

      call pgbbuf
      call pgeras

      ! Draw background box edges:
!      call pgsci(1)
      call drawBox(box, edgeIsBkg, edgeCI)

      ! Draw points:
      call pgsci(pointsCI)
      if (withGrid) then
        if (xsize .gt. 1) then
          do yi = 1, ysize
            call pgline(xsize, normalisedPoints(:, yi)%proj(1)&
            , normalisedPoints(:, yi)%proj(2))
          end do
        end if
        if (ysize .gt. 1) then
          do xi = 1, xsize
            call pgline(ysize, normalisedPoints(xi, :)%proj(1)&
            , normalisedPoints(xi, :)%proj(2))
          end do
        end if

      else
        if (xsize .gt. ysize) then
          do yi = 1, ysize
            call pgpt(xsize, normalisedPoints(:, yi)%proj(1)&
            , normalisedPoints(:, yi)%proj(2), 1)
          end do

        else
          do xi = 1, xsize
            call pgpt(ysize, normalisedPoints(xi, :)%proj(1)&
            , normalisedPoints(xi, :)%proj(2), 1)
          end do
        end if
      end if

      ! Draw foreground edges:
!      call pgsci(1)
      call drawBox(box, .not. edgeIsBkg, edgeCI)

      call pgebuf
    end do

    call pgclos

  end subroutine rotatingDisplayKernel

  !---------------------------------------------------------------------
  subroutine decideBkgEdges(obslocation, boxMin, boxMax, edgeIsBkg)
  !---------------------------------------------------------------------

    implicit none

    real, dimension(3), intent(in) :: obslocation, boxMin, boxMax
    logical, intent(out) :: edgeIsBkg(3, 2, 2)

    !-------------------------------------------------------------------
    ! Start

    edgeIsBkg = .true. ! array op, default

    if (obslocation(1) .lt. boxMin(1)) then
      edgeIsBkg(2, 1, :) = .false.
      edgeIsBkg(3, 1, :) = .false.

    else if (obslocation(1) .gt. boxMax(1)) then
      edgeIsBkg(2, 2, :) = .false.
      edgeIsBkg(3, 2, :) = .false.
    end if

    if (obslocation(2) .lt. boxMin(2)) then
      edgeIsBkg(1, 1, :) = .false.
      edgeIsBkg(3, :, 1) = .false.

    else if (obslocation(2) .gt. boxMax(2)) then
      edgeIsBkg(1, 2, :) = .false.
      edgeIsBkg(3, :, 2) = .false.
    end if

    if (obslocation(3) .lt. boxMin(3)) then
      edgeIsBkg(1, :, 1) = .false.
      edgeIsBkg(2, :, 1) = .false.

    else if (obslocation(3) .gt. boxMax(3)) then
      edgeIsBkg(1, :, 2) = .false.
      edgeIsBkg(2, :, 2) = .false.
    end if

  end subroutine decideBkgEdges

  !---------------------------------------------------------------------
  subroutine drawBox(box, drawEdge, edgeCI)
  !---------------------------------------------------------------------

    implicit none

    type(pointType), intent(in) :: box(2,2,2)
    logical, intent(in) :: drawEdge(3, 2, 2)
    integer, intent(in) :: edgeCI(3, 2, 2)

    integer :: xi, yi, zi

    !-------------------------------------------------------------------
    ! Start

    do yi = 1, 2
      do zi = 1, 2
        if (drawEdge(1, yi, zi)) then
          call pgsci(edgeCI(1, yi, zi))
          call pgmove(box(1, yi, zi)%proj(1), box(1, yi, zi)%proj(2))
          call pgdraw(box(2, yi, zi)%proj(1), box(2, yi, zi)%proj(2))
        end if
      end do
    end do
    do xi = 1, 2
      do zi = 1, 2
        if (drawEdge(2, xi, zi)) then
          call pgsci(edgeCI(2, xi, zi))
          call pgmove(box(xi, 1, zi)%proj(1), box(xi, 1, zi)%proj(2))
          call pgdraw(box(xi, 2, zi)%proj(1), box(xi, 2, zi)%proj(2))
        end if
      end do
    end do
    do xi = 1, 2
      do yi = 1, 2
        if (drawEdge(3, xi, yi)) then
          call pgsci(edgeCI(3, xi, yi))
          call pgmove(box(xi, yi, 1)%proj(1), box(xi, yi, 1)%proj(2))
          call pgdraw(box(xi, yi, 2)%proj(1), box(xi, yi, 2)%proj(2))
        end if
      end do
    end do

  end subroutine drawBox

  !---------------------------------------------------------------------
  subroutine stereoDisplayVector(points, numStepsPerQuadrant, obsThetaDeg&
  , obsPhiDeg, ranges, device, withGrid, eyesCrossed, overlap, status)
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    real,         intent(in)            :: points(:, :)
    integer,      intent(in),  optional :: numStepsPerQuadrant
    real,         intent(in),  optional :: obsThetaDeg,&
                                           obsPhiDeg,&
                                           ranges(3, 2),&
                                           overlap
    character(*), intent(in),  optional :: device
    logical,      intent(in),  optional :: withGrid,&
                                           eyesCrossed
    integer,      intent(out), optional :: status

    real ::&
      localObsThetaDeg,&
      localRanges(3, 2),&
      localPoints(size(points, 1), 1, size(points, 2)),&
      localObsPhiDeg,&
      localOverlap
    character(255) ::&
      localDevice
    integer ::&
      localStatus,&
      coord,&
      localNumStepsPerQuadrant
    logical ::&
      localWithGrid,&
      localEyesCrossed

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    localPoints(:, 1, :) = points

    call initDefaultObs

    if (present(numStepsPerQuadrant)) then
      localNumStepsPerQuadrant = numStepsPerQuadrant
    else
      localNumStepsPerQuadrant = 45
    end if

    if (present(obsThetaDeg)) then
      localObsThetaDeg = obsThetaDeg
      do while (localObsThetaDeg .lt. 0.0)
        localObsThetaDeg = localObsThetaDeg + 360.0
      end do
      do while (localObsThetaDeg .ge. 360.0)
        localObsThetaDeg = localObsThetaDeg - 360.0
      end do
    else
      localObsThetaDeg = defaultObs%locTheta / myPiOn180
    end if

    if (present(obsPhiDeg)) then
      localObsPhiDeg = obsPhiDeg
    else
      localObsPhiDeg = defaultObs%locPhi / myPiOn180
    end if

    if (present(ranges)) then
      localRanges = ranges
    else
      do coord = 1, 3
        localRanges(coord, 1) = minval(points(:, coord))
        localRanges(coord, 2) = maxval(points(:, coord))
      end do
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(withGrid)) then
      localWithGrid = withGrid
    else
      localWithGrid = .false.
    end if

    if (present(eyesCrossed)) then
      localEyesCrossed = eyesCrossed
    else
      localEyesCrossed = .false.
    end if

    if (present(overlap)) then
      localOverlap = overlap
    else
      localOverlap = 0.0
    end if

    call stereoDisplayKernel(localPoints&
    , localObsThetaDeg, localObsPhiDeg, localRanges, localDevice&
    , localWithGrid, localEyesCrossed, localOverlap, localStatus)

    if(present(status)) status = localStatus

  end subroutine stereoDisplayVector

  !---------------------------------------------------------------------
  subroutine stereoDisplayArray(points, numStepsPerQuadrant, obsThetaDeg&
  , obsPhiDeg, ranges, device, withGrid, eyesCrossed, overlap, status)
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    real,         intent(in)            :: points(:, :, :)
    integer,      intent(in),  optional :: numStepsPerQuadrant
    real,         intent(in),  optional :: obsThetaDeg,&
                                           obsPhiDeg,&
                                           ranges(3, 2),&
                                           overlap
    character(*), intent(in),  optional :: device
    logical,      intent(in),  optional :: withGrid,&
                                           eyesCrossed
    integer,      intent(out), optional :: status

    real ::&
      localObsThetaDeg,&
      localRanges(3, 2),&
      localPoints(size(points, 1), size(points, 2), size(points, 3)),&
      localObsPhiDeg,&
      localOverlap
    character(255) ::&
      localDevice
    integer ::&
      localStatus,&
      coord,&
      localNumStepsPerQuadrant
    logical ::&
      localWithGrid,&
      localEyesCrossed

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0 ! default

    localPoints = points

    call initDefaultObs

    if (present(numStepsPerQuadrant)) then
      localNumStepsPerQuadrant = numStepsPerQuadrant
    else
      localNumStepsPerQuadrant = 45
    end if

    if (present(obsThetaDeg)) then
      localObsThetaDeg = obsThetaDeg
      do while (localObsThetaDeg .lt. 0.0)
        localObsThetaDeg = localObsThetaDeg + 360.0
      end do
      do while (localObsThetaDeg .ge. 360.0)
        localObsThetaDeg = localObsThetaDeg - 360.0
      end do
    else
      localObsThetaDeg = defaultObs%locTheta / myPiOn180
    end if

    if (present(obsPhiDeg)) then
      localObsPhiDeg = obsPhiDeg
    else
      localObsPhiDeg = defaultObs%locPhi / myPiOn180
    end if

    if (present(ranges)) then
      localRanges = ranges
    else
      do coord = 1, 3
        localRanges(coord, 1) = minval(points(:, :, coord))
        localRanges(coord, 2) = maxval(points(:, :, coord))
      end do
    end if

    if (present(device)) then
      localDevice = device
    else
      localDevice = defaultDevice
    end if

    if (present(withGrid)) then
      localWithGrid = withGrid
    else
      localWithGrid = .false.
    end if

    if (present(eyesCrossed)) then
      localEyesCrossed = eyesCrossed
    else
      localEyesCrossed = .false.
    end if

    if (present(overlap)) then
      localOverlap = overlap
    else
      localOverlap = 0.0
    end if

    call stereoDisplayKernel(localPoints&
    , localObsThetaDeg, localObsPhiDeg, localRanges, localDevice&
    , localWithGrid, localEyesCrossed, localOverlap, localStatus)

    if(present(status)) status = localStatus

  end subroutine stereoDisplayArray

  !---------------------------------------------------------------------
  subroutine stereoDisplayKernel(points, obsThetaDeg&
  , obsPhiDeg, ranges, device, withGrid, eyesCrossed, overlap, status)
  !---------------------------------------------------------------------
!*** sort out status numbers.

    implicit none

    !-------------------------------------------------------------------
    ! Start

    real,         intent(in)  :: points(:, :, :)
    real,         intent(in)  :: obsThetaDeg,&
                                 obsPhiDeg,&
                                 ranges(3, 2),&
                                 overlap
    character(*), intent(in)  :: device
    logical,      intent(in)  :: withGrid,&
                                 eyesCrossed
    integer,      intent(out) :: status

    real ::&
      boxMin(3), boxMax(3), radius2, maxRadius2, eyeSeparation,&
      projMaxDiameter, eyeSepAngle
    integer :: pgopen, xsize, ysize, coord, foreignStatus, deviceId, xi, yi,&
      zi, i, edgeCI(3, 2, 2), pointsCI
    type(obsType) :: leftObs, riteObs, myObs
! has real :: location(3), locRadius, locTheta, locPhi, viewFrame(3, 3), viewTheta, viewPhi, viewTwist, projOffsets(2), projScale
    type(pointType) ::&
      box(2, 2, 2),& ! the bounding box of the displayed surface.
      normalisedPoints(size(points, 1), size(points, 2))
! has real :: solid(3), proj(2)
    logical :: edgeIsBkg(3, 2, 2)

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default

    xsize = size(points, 1)
    ysize = size(points, 2)
    if (size(points, 3) .ne. 3) then
      status = -1
  return
    end if

    edgeCI = 1 ! array op, default.
    edgeCI(1, 1, 1) = 2
    edgeCI(2, 1, 1) = 3
    edgeCI(3, 1, 1) = 4
    pointsCI = 5

    leftObs%locRadius = defaultObs%locRadius
    riteObs%locRadius = defaultObs%locRadius
    eyeSeparation = 0.4
    eyeSepAngle = atan2(eyeSeparation / 2.0, defaultObs%locRadius)

    leftObs%locTheta = obsThetaDeg * myPiOn180 - eyeSepAngle
    riteObs%locTheta = obsThetaDeg * myPiOn180 + eyeSepAngle
    if (obsPhiDeg .lt. -90.0 .or. obsPhiDeg .gt. 90.0) then
      status = -1
  return
    end if
    leftObs%locPhi = obsPhiDeg * myPiOn180
    riteObs%locPhi = obsPhiDeg * myPiOn180

    boxMin = -0.5
    boxMax = 0.5

    box(1, 1:2, 1:2)%solid(1) = boxMin(1)
    box(2, 1:2, 1:2)%solid(1) = boxMax(1)
    box(1:2, 1, 1:2)%solid(2) = boxMin(2)
    box(1:2, 2, 1:2)%solid(2) = boxMax(2)
    box(1:2, 1:2, 1)%solid(3) = boxMin(3)
    box(1:2, 1:2, 2)%solid(3) = boxMax(3)

    call fillOutObs(leftObs)
    call fillOutObs(riteObs)
    leftObs%projOffsets = 0.0
    riteObs%projOffsets = 0.0
    ! Find max radius of box corners:
    maxRadius2&
    = box(1, 1, 1)%solid(1) * box(1, 1, 1)%solid(1)&
    + box(1, 1, 1)%solid(2) * box(1, 1, 1)%solid(2)&
    + box(1, 1, 1)%solid(3) * box(1, 1, 1)%solid(3)
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          radius2&
          = box(xi, yi, zi)%solid(1) * box(xi, yi, zi)%solid(1)&
          + box(xi, yi, zi)%solid(2) * box(xi, yi, zi)%solid(2)&
          + box(xi, yi, zi)%solid(3) * box(xi, yi, zi)%solid(3)
          maxRadius2 = max(maxRadius2, radius2)
        end do
      end do
    end do
    if (leftObs%locRadius .le. maxRadius2) then
      status = -2
  return
    end if

    projMaxDiameter = 2.0 * sqrt(maxRadius2&
    / (leftObs%locRadius * leftObs%locRadius - maxRadius2))
    leftObs%projScale = 1.0 / projMaxDiameter
    riteObs%projScale = 1.0 / projMaxDiameter

    do coord = 1, 3
      if (ranges(coord, 2) - ranges(coord, 1) .le. 0.0) then
!****** set warning status
        normalisedPoints%solid(coord) = 0.5
      else
        normalisedPoints%solid(coord)&
        = (points(:, :, coord) - ranges(coord, 1))&
        / (ranges(coord, 2) - ranges(coord, 1))
      end if

      normalisedPoints%solid(coord)&
      = boxMin(coord) * (1.0 - normalisedPoints%solid(coord))&
      + boxMax(coord) * normalisedPoints%solid(coord)
    end do

    deviceId = pgopen(trim(device))
    call findClrLimits(foreignStatus)
!***** integrate foreignStatus and status
  if (status .lt. 0) return

    call pgscr(pointsCI, 0.5, 0.5, 0.5)
    call pgsfs(2) ! unfilled

!.......................................................................
! Draw LH cube:

    if (eyesCrossed) then
      myObs = leftObs
    else
      myObs = riteObs
    end if

    call pgsvp(0.0, 0.5, 0.0, 0.999)
    call pgwnad(-0.5, 0.5, -0.5, 0.5)

    ! Calculate projected point positions:
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          call project_new(box(xi, yi, zi)%solid, myObs&
          , box(xi, yi, zi)%proj, foreignStatus)
!***** integrate foreignStatus and localStatus
  if (status .lt. 0) return
        end do
      end do
    end do
    do xi = 1, xsize
      do yi = 1, ysize
        call project_new(normalisedPoints(xi, yi)%solid, myObs&
        , normalisedPoints(xi, yi)%proj, foreignStatus)
!***** integrate foreignStatus and status
  if (status .lt. 0) return
      end do
    end do

    ! Decide which are bkg and which frg box edges:
    call decideBkgEdges(myObs%location, boxMin, boxMax, edgeIsBkg)

    ! Draw background box edges:
    call drawBox(box, edgeIsBkg, edgeCI)

    ! Draw points:
    call pgsci(pointsCI)
    if (withGrid) then
      if (xsize .gt. 1) then
        do yi = 1, ysize
          call pgline(xsize, normalisedPoints(:, yi)%proj(1)&
          , normalisedPoints(:, yi)%proj(2))
        end do
      end if
      if (ysize .gt. 1) then
        do xi = 1, xsize
          call pgline(ysize, normalisedPoints(xi, :)%proj(1)&
          , normalisedPoints(xi, :)%proj(2))
        end do
      end if

    else
      if (xsize .gt. ysize) then
        do yi = 1, ysize
          call pgpt(xsize, normalisedPoints(:, yi)%proj(1)&
          , normalisedPoints(:, yi)%proj(2), 1)
        end do

      else
        do xi = 1, xsize
          call pgpt(ysize, normalisedPoints(xi, :)%proj(1)&
          , normalisedPoints(xi, :)%proj(2), 1)
        end do
      end if
    end if

    ! Draw foreground edges:
    call drawBox(box, .not. edgeIsBkg, edgeCI)

!.......................................................................
! Draw RH cube:

    if (eyesCrossed) then
      myObs = riteObs
    else
      myObs = leftObs
    end if

    call pgsvp(0.5, 0.999, 0.0, 0.999)
    call pgwnad(-0.5, 0.5, -0.5, 0.5)

    ! Calculate projected point positions:
    do xi = 1, 2
      do yi = 1, 2
        do zi = 1, 2
          call project_new(box(xi, yi, zi)%solid, myObs&
          , box(xi, yi, zi)%proj, foreignStatus)
!***** integrate foreignStatus and localStatus
  if (status .lt. 0) return
        end do
      end do
    end do
    do xi = 1, xsize
      do yi = 1, ysize
        call project_new(normalisedPoints(xi, yi)%solid, myObs&
        , normalisedPoints(xi, yi)%proj, foreignStatus)
!***** integrate foreignStatus and status
  if (status .lt. 0) return
      end do
    end do

    ! Decide which are bkg and which frg box edges:
    call decideBkgEdges(myObs%location, boxMin, boxMax, edgeIsBkg)

    ! Draw background box edges:
    call drawBox(box, edgeIsBkg, edgeCI)

    ! Draw points:
    call pgsci(pointsCI)
    if (withGrid) then
      if (xsize .gt. 1) then
        do yi = 1, ysize
          call pgline(xsize, normalisedPoints(:, yi)%proj(1)&
          , normalisedPoints(:, yi)%proj(2))
        end do
      end if
      if (ysize .gt. 1) then
        do xi = 1, xsize
          call pgline(ysize, normalisedPoints(xi, :)%proj(1)&
          , normalisedPoints(xi, :)%proj(2))
        end do
      end if

    else
      if (xsize .gt. ysize) then
        do yi = 1, ysize
          call pgpt(xsize, normalisedPoints(:, yi)%proj(1)&
          , normalisedPoints(:, yi)%proj(2), 1)
        end do

      else
        do xi = 1, xsize
          call pgpt(ysize, normalisedPoints(xi, :)%proj(1)&
          , normalisedPoints(xi, :)%proj(2), 1)
        end do
      end if
    end if

    ! Draw foreground edges:
    call drawBox(box, .not. edgeIsBkg, edgeCI)

    call pgclos

  end subroutine stereoDisplayKernel

  !---------------------------------------------------------------------
  subroutine drawBox_old(box, drawEdge)
  !---------------------------------------------------------------------

    implicit none

    type(pointType), intent(in) :: box(2,2,2)
    logical, intent(in) :: drawEdge(12)

    !-------------------------------------------------------------------
    ! Start

    if (drawEdge(1)) then
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
    end if
    if (drawEdge(2)) then
      call pgmove(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
      call pgdraw(box(2, 1, 2)%proj(1), box(2, 1, 2)%proj(2))
    end if
    if (drawEdge(3)) then
      call pgmove(box(2, 1, 2)%proj(1), box(2, 1, 2)%proj(2))
      call pgdraw(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))
    end if
    if (drawEdge(4)) then
      call pgmove(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))
      call pgdraw(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
    end if

    if (drawEdge(5)) then
      call pgmove(box(1, 1, 1)%proj(1), box(1, 1, 1)%proj(2))
      call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
    end if
    if (drawEdge(6)) then
      call pgmove(box(2, 1, 1)%proj(1), box(2, 1, 1)%proj(2))
      call pgdraw(box(2, 2, 1)%proj(1), box(2, 2, 1)%proj(2))
    end if
    if (drawEdge(7)) then
      call pgmove(box(2, 1, 2)%proj(1), box(2, 1, 2)%proj(2))
      call pgdraw(box(2, 2, 2)%proj(1), box(2, 2, 2)%proj(2))
    end if
    if (drawEdge(8)) then
      call pgmove(box(1, 1, 2)%proj(1), box(1, 1, 2)%proj(2))
      call pgdraw(box(1, 2, 2)%proj(1), box(1, 2, 2)%proj(2))
    end if

    if (drawEdge(9)) then
      call pgmove(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
      call pgdraw(box(2, 2, 1)%proj(1), box(2, 2, 1)%proj(2))
    end if
    if (drawEdge(10)) then
      call pgmove(box(2, 2, 1)%proj(1), box(2, 2, 1)%proj(2))
      call pgdraw(box(2, 2, 2)%proj(1), box(2, 2, 2)%proj(2))
    end if
    if (drawEdge(11)) then
      call pgmove(box(2, 2, 2)%proj(1), box(2, 2, 2)%proj(2))
      call pgdraw(box(1, 2, 2)%proj(1), box(1, 2, 2)%proj(2))
    end if
    if (drawEdge(12)) then
      call pgmove(box(1, 2, 2)%proj(1), box(1, 2, 2)%proj(2))
      call pgdraw(box(1, 2, 1)%proj(1), box(1, 2, 1)%proj(2))
    end if

  end subroutine drawBox_old

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module plotutils_mod
